from typing import Iterable, List, Mapping, Optional, Set, Union
from collections import namedtuple
from warnings import warn

import networkx as nx
from IPython import get_ipython
from IPython.display import Markdown
from graphviz import Digraph
from rich.table import Table

from flowchem.units import flowchem_ureg
from flowchem.components.stdlib import Component, Tube, Valve, Vessel

Connection = namedtuple("Connection", ["from_component", "to_component", "tube"])


class Apparatus(object):
    """
    A unique network of components.

    Arguments:
    - `name`: The name of the apparatus. This should be short and sweet.
    - `description`: A description of the apparatus. Can be as long and wordy as you want.

    Attributes:
    - `components`: A set of components that make up the apparatus.
    - `name`: The name of the apparatus.
    - `description`: A description of the apparatus.
    - `network`: Network edge list as list of Connection tuples i.e. `(from_component, to_component, tube)`.
    """

    _id_counter = 0

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None):
        self.network: List[Connection] = []
        self.components: Set[Component] = set()
        # if given a name, then name the apparatus, else default to a sequential name
        if name is not None:
            self.name = name
        else:
            self.name = "Apparatus_" + str(Apparatus._id_counter)
            Apparatus._id_counter += 1
        self.description = description

    def __repr__(self):
        return f"<Apparatus {self.name}>"

    def __str__(self):
        return f"Apparatus {self.name}"

    def __getitem__(self, item):
        # when you pass a class
        if isinstance(item, type):
            return [
                component
                for component in self.components
                if isinstance(component, item)
            ]
        elif isinstance(item, str):
            try:
                return [
                    component for component in self.components if component.name == item
                ][0]
            except IndexError:
                raise KeyError(f"No component named '{item}' in {repr(self)}.")

        # a shorthand way to check if a component is in the apparatus
        elif isinstance(item, Component):
            if item in self.components:
                return item
            else:
                raise KeyError(f"{repr(item)} is not in {repr(self)}.")

    def _add_single(
        self, from_component: Component, to_component: Component, tube: Tube
    ) -> None:
        """Adds a single connection to the apparatus.

        For args, see add().
        """
        assert isinstance(from_component, Component), "From component must be an instance of Component"
        assert isinstance(to_component, Component), "To component must be an instance of Component"
        assert isinstance(tube, Tube), "Tube must be an instance of Tube"

        # check for duplicate names
        try:
            if self[from_component.name] is not from_component:
                raise ValueError(f"Component {from_component} has duplicated name")
        except KeyError:
            pass
        try:
            if self[to_component.name] is not to_component:
                raise ValueError(f"Component {to_component} has duplicated name")
        except KeyError:
            pass

        if (
            Connection(
                from_component=from_component, to_component=to_component, tube=tube
            )
            in self.network
        ):
            warn(
                f"Duplicate connection from {from_component} to {to_component} omitted."
            )
            return

        self.network.append(
            Connection(
                from_component=from_component, to_component=to_component, tube=tube
            )
        )
        self.components.update([from_component, to_component])

    def add(
        self,
        from_component: Union[Component, Iterable[Component]],
        to_component: Union[Component, Iterable[Component]],
        tube: Tube,
    ) -> None:
        """
        Adds connections to the apparatus.
        If both `from_component` and `to_component` are iterables, then their Cartesian product will be added to the apparatus.

        Arguments:
        - `from_component`: The `Component` from which the flow is originating. If an iterable, all items in the iterable will be connected to the same component.
        - `to_component`: The `Component` where the flow is going. If an iterable, all items in the iterable will be connected to the same component.
        - `tube`: `Tube` that connects the components.

        Raises:
        - `ValueError`: When the connection being added is invalid.
        """

        # the cartesian product
        if isinstance(from_component, Iterable) and isinstance(to_component, Iterable):
            for _from_component in from_component:
                for _to_component in to_component:
                    self._add_single(_from_component, _to_component, tube)

        # multiple from components, one to component
        elif isinstance(from_component, Iterable) and not isinstance(
            to_component, Iterable
        ):
            for _from_component in from_component:
                self._add_single(_from_component, to_component, tube)

        # multiple to components, one from component
        elif isinstance(to_component, Iterable) and not isinstance(
            from_component, Iterable
        ):
            for _to_component in to_component:
                self._add_single(from_component, _to_component, tube)

        # one to and one from component
        elif not isinstance(to_component, Iterable) and not isinstance(
            from_component, Iterable
        ):
            self._add_single(from_component, to_component, tube)

    def visualize(
        self,
        title: Union[bool, str] = True,
        label_tubes: bool = False,
        describe_vessels: bool = False,
        rankdir: str = "LR",
        node_attr: Optional[Mapping[str, str]] = None,
        edge_attr: Optional[Mapping[str, str]] = None,
        graph_attr: Optional[Mapping[str, str]] = None,
        file_format: str = "pdf",
        filename: Optional[str] = None,
        **kwargs,
    ) -> Optional[Digraph]:
        """
        Generates a visualization of an apparatus's network graph.

        For full list of acceptable Graphviz attributes, see [the graphviz.org docs](http://www.graphviz.org/doc/info/attrs.html) and [its Python API's docs](http://graphviz.readthedocs.io/en/stable/manual.html#attributes).

        Arguments:
        - `describe_vessels`: Whether to display the names or content descriptions of `Vessel` components.
        - `edge_attr`: Controls the appearance of the edges (tubes) of the Apparatus. Must be of the form `{"attribute": "value"}`.
        - `file_format`: The output format of the graph, either "pdf" or "png".
        - `filename`: The name of the output file. Defaults to the name of the apparatus.
        - `graph_attr`: Controls the appearance of the Apparatus. Must be of the form `{"attribute": "value"}`. To get orthogonal splines (*i.e.* edges with sharp corners), pass `splines="ortho"`. To increase the separation between components, set `nodesep = "0.5"` or similar.
        - `label_tubes`: Whether to label the tubes between components with the length, inner diameter, and outer diameter.
        - `node_attr`: Controls the appearance of the nodes (components) of the Apparatus. Must be of the form `{"attribute": "value"}`.
        - `rankdir`: The direction of the graph. Use `LR` for left to right and `TD` for top down.
        - `title`: Whether to show the title in the output. Defaults to True. If a string, the title to use for the output.
        """
        f = Digraph(
            name=self.name,
            node_attr=node_attr,
            edge_attr=edge_attr,
            graph_attr=graph_attr if graph_attr is not None else {},
            format=file_format,
            filename=filename,
            **kwargs,
        )

        # go from left to right adding components and their tubing connections
        f.attr(rankdir=rankdir)

        for component in sorted(list(self.components), key=lambda x: x.name):
            f.attr("node", shape=component._visualization_shape)
            f.node(
                component.description
                if isinstance(component, Vessel) and describe_vessels
                else component.name
            )

        for c in self.network:
            tube = c.tube
            tube_label = (
                f"Length {tube.length}\nID {tube.ID}\nOD {tube.OD}"
                if label_tubes
                else ""
            )
            f.edge(
                c.from_component.description
                if isinstance(c.from_component, Vessel) and describe_vessels
                else c.from_component.name,
                c.to_component.description
                if isinstance(c.to_component, Vessel) and describe_vessels
                else c.to_component.name,
                label=tube_label,
            )

        # show the title of the graph
        if title:
            title = title if isinstance(title, str) else self.name
            f.attr(label=title)

        if get_ipython():
            return f
        else:
            f.view(cleanup=True)
            return None

    def summarize(self) -> None:
        """
        Prints a summary table of the apparatus.

        Returns:
        - In Jupyter, a nice HTML table. Otherwise, the output is printed to the terminal.
        """

        # Components table
        components_table = Table(title="Components")

        # Columns: Name, Type
        components_table.add_column("Name")
        components_table.add_column("Type")

        # Fill rows
        for component in sorted(self.components, key=lambda x: x.__class__.__name__):
            components_table.add_row(component.name, component.__class__.__name__)

        # Tubing table
        tubing_table = Table("From", "To", "Length", "I.D.", "O.D.", "Volume", "Material", title="Tubing")

        # store and calculate the computed totals for tubing
        total_length = 0 * flowchem_ureg.mm
        total_volume = 0 * flowchem_ureg.ml
        for connection in self.network:
            total_length += connection.tube.length
            total_volume += connection.tube.volume

            tubing_table.add_row(
                connection.from_component.name,
                connection.to_component.name,
                round(connection.tube.length, 4),
                round(connection.tube.ID, 4),
                round(connection.tube.OD, 4),
                round(connection.tube.volume.to("ml"), 4),
                connection.tube.material)

        tubing_table.add_row(
            "Total",
            "n/a",
            round(total_length, 4),
            "n/a",
            "n/a",
            round(total_volume.to("ml"), 4),
            "n/a")

        # Print tables
        from rich.console import Console
        console = Console()
        console.print(components_table)
        console.print(tubing_table)

    def _validate(self) -> bool:
        """
        Checks that the apparatus is valid.

        ::: tip
        Calling this function yourself is likely unnecessary because the `Protocol` class calls it upon instantiation.
        :::

        Returns:
        - Whether the apparatus is valid.
        """

        # make sure that all of the components are connected
        G = nx.Graph()  # convert the network to an undirected NetworkX graph
        G.add_edges_from([(c.from_component, c.to_component) for c in self.network])
        if not nx.is_connected(G):
            warn("Not all components connected.")
            return False

        # valve checking
        for valve in self[Valve]:

            # ensure that valve's mapping components are part of apparatus
            if isinstance(valve.mapping, Mapping):
                for component in valve.mapping.keys():
                    if component not in self.components:
                        warn(
                            f"Invalid mapping for Valve {valve}. "
                            f"{component} has not been added to {self.name}"
                        )
                        return False

            # TODO: make this check work again with SISO, SIMO, MISO, and MIMO valves.
            # # no more than one output from a valve (might have to change this)
            # if len([x for x in self.network if x.from_component is valve]) != 1:
            #     warn(f"Valve {valve} has multiple outputs.")
            #     return False
            #
            # make sure valve's mapping is complete
            # non_mapped_components = [
            #     connection.from_component
            #     for connection in self.network
            #     if connection.to_component == valve
            #     and valve.mapping.get(connection.from_component.name) is None
            # ]
            # if non_mapped_components:
            #     warn(
            #         f"Valve {valve} has incomplete mapping."
            #         f" No mapping for {non_mapped_components}"
            #     )
            #     return False

        return True

    def describe(self) -> Union[str, Markdown]:
        """
        Generates a human-readable description of the apparatus.

        Returns:
        - A description of apparatus. When in Jupyter, this string is wrapped in a `IPython.display.Markdown` object for nicer display.

        Raises:
        - `RuntimeError`: When a component cannot be described.
        """

        def _description(element, capitalize=False):
            """takes a component and converts it to a string description"""
            if issubclass(element.__class__, Vessel):
                return f"{'A' if capitalize else 'a'} vessel containing {element.description}"
            elif issubclass(element.__class__, Component):
                return element.__class__.__name__ + " " + element.name
            else:
                raise RuntimeError(
                    f"{element} cannot be described."
                    " If you're seeing this message, something *very* wrong has happened."
                )

        result = ""

        # iterate over the network and describe the connections
        for connection in self.network:
            from_component, to_component, tube = (
                _description(connection.from_component, capitalize=True),
                _description(connection.to_component),
                connection.tube,
            )
            result += (
                f"{from_component} was connected to"
                f" {to_component} using {connection[2].material}"
                f" tubing (length {tube.length}, ID {tube.ID}, OD {tube.OD}). "
            )
        if get_ipython():
            return Markdown(result)
        return result



from typing import Optional

from flowchem.components.stdlib import Component


class Vessel(Component):
    """
    A generic vessel.

    Arguments:
    - `description`: The contents of the vessel.
    - `name`: The name of the vessel, if different from the description.

    Attributes:
    - `description`: The contents of the vessel.
    - `name`: The name of the vessel, if different from the description.
    """

    def __init__(self, description: Optional[str] = None, name: Optional[str] = None):
        super().__init__(name=name)
        self.description = description
        self._visualization_shape = "cylinder"



""" NMR-spectrum object represents an NMR spectrum.  """
import nmrglue as ng
from pathlib import Path
import matplotlib.pyplot as plt


class NMRSpectrum:
    """General spectrum object, instantiated from Spinsolve folder w/ experimental results."""

    def __init__(self, location: Path):
        self.dic, self.raw_data = ng.spinsolve.read(dir=location.as_posix())
        self.processed_data = None

    @property
    def uc(self):
        """

        Returns:

        """
        data = self.processed_data if self.processed_data is not None else self.raw_data
        return ng.spinsolve.make_uc(self.dic, data)

    def process(self):
        """Basic spectrum processing. Application-specific processing suggested."""
        # Zerofill
        self.processed_data = ng.proc_base.zf_auto(
            ng.proc_base.zf_double(self.raw_data, 1)
        )

        # FT
        self.processed_data = ng.proc_base.fft(self.processed_data)

        # Autophase
        self.processed_data = ng.proc_autophase.autops(
            self.processed_data, "acme", disp=False
        )

        # Delete imaginary
        self.processed_data = ng.proc_base.di(self.processed_data)

    def plot(self, ppm_range=(8, 0)):
        """Returns spectrum as matplotlib figure"""
        if self.processed_data is None:
            self.process()

        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(self.uc.ppm_scale(), self.processed_data)
        plt.xlim(ppm_range)  # plot as we are used to, from positive to negative
        return fig



""" Async implementation of FlowIR """

import datetime
import warnings
from loguru import logger
from typing import List, Optional

from flowchem.exceptions import DeviceError

import asyncio
from asyncua import ua, Client

from flowchem.components.devices.MettlerToledo.iCIR_common import (
    IRSpectrum,
    iCIR_spectrometer,
    ProbeInfo,
)


class FlowIR(iCIR_spectrometer):
    """
    Object to interact with the iCIR software controlling the FlowIR and ReactIR.
    """

    def __init__(self, url: str = None):
        """
        Initiate connection with OPC UA server.
        Intended to be used as context-manager!
        """
        # Default (local) url if none provided
        if url is None:
            url = "opc.tcp://localhost:62552/iCOpcUaServer"

        self.opcua = Client(url)
        self.version = None

    async def initialize(self):
        """Initialize and check connection"""
        try:
            await self.opcua.connect()
        except asyncio.TimeoutError:
            raise DeviceError(
                f"Could not connect to FlowIR on {self.opcua.server_url}!"
            )
        await self.check_version()
        logger.debug("FlowIR initialized!")

    async def check_version(self):
        """Check if iCIR is installed and open and if the version is supported."""
        try:
            self.version = await self.opcua.get_node(
                self.SOFTWARE_VERSION
            ).get_value()  # "7.1.91.0"
            if self.version not in self._supported_versions:
                warnings.warn(
                    f"The current version of iCIR [self.version] has not been tested!"
                    f"Pleas use one of the supported versions: {self._supported_versions}"
                )
        except ua.UaStatusCodeError as e:  # iCIR app closed
            raise DeviceError(
                "iCIR app not installed or closed or no instrument available!"
            ) from e

    # noinspection PyPep8Naming
    async def is_iCIR_connected(self) -> bool:
        """Check connection with instrument"""
        return await self.opcua.get_node(self.CONNECTION_STATUS).get_value()

    async def probe_info(self) -> ProbeInfo:
        """Return FlowIR probe information"""
        probe_info = await self.opcua.get_node(self.PROBE_DESCRIPTION).get_value()
        return self.parse_probe_info(probe_info)

    async def probe_status(self):
        """Returns current probe status"""
        return await self.opcua.get_node(self.PROBE_STATUS).get_value()

    async def is_running(self) -> bool:
        """Is the probe currently measuring?"""
        return await self.probe_status() == "Running"

    async def last_sample_time(self) -> datetime.datetime:
        """Returns date/time of latest scan"""
        return await self.opcua.get_node(self.LAST_SAMPLE_TIME).get_value()

    async def sample_count(self) -> Optional[int]:
        """Sample count (integer autoincrement) watch for changes to ensure latest spectrum is recent"""
        return await self.opcua.get_node(self.SAMPLE_COUNT).get_value()

    @staticmethod
    async def _wavenumber_from_spectrum_node(node) -> List[float]:
        """Gets the X-axis value of a spectrum. This is necessary as they change e.g. with resolution."""
        node_property = await node.get_properties()
        x_axis = await node_property[0].get_value()
        return x_axis.AxisSteps

    @staticmethod
    async def spectrum_from_node(node) -> IRSpectrum:
        """Given a Spectrum node returns it as IRSpectrum"""
        try:
            intensity = await node.get_value()
            wavenumber = await FlowIR._wavenumber_from_spectrum_node(node)
            return IRSpectrum(wavenumber=wavenumber, intensity=intensity)

        except ua.uaerrors.BadOutOfService:
            return IRSpectrum(wavenumber=[], intensity=[])

    async def last_spectrum_treated(self) -> IRSpectrum:
        """Returns an IRSpectrum element for the last acquisition"""
        return await FlowIR.spectrum_from_node(
            self.opcua.get_node(self.SPECTRA_TREATED)
        )

    async def last_spectrum_raw(self) -> IRSpectrum:
        """RAW result latest scan"""
        return await FlowIR.spectrum_from_node(self.opcua.get_node(self.SPECTRA_RAW))

    async def last_spectrum_background(self) -> IRSpectrum:
        """RAW result latest scan"""
        return await FlowIR.spectrum_from_node(
            self.opcua.get_node(self.SPECTRA_BACKGROUND)
        )

    async def start_experiment(
        self, template: str, name: str = "Unnamed flowchem exp."
    ):
        """Starts an experiment on iCIR

        Args:
            template: name of the experiment template, should be in the right folder on the PC running iCIR
            name: experiment name.
        """
        template = FlowIR._normalize_template_name(template)
        if self.is_local() and FlowIR.is_template_name_valid(template) is False:
            raise DeviceError(
                f"Cannot start template {template}: name not valid! Check if is in: "
                r"C:\ProgramData\METTLER TOLEDO\iC OPC UA Server\1.2\Templates"
            )
        if await self.is_running():
            warnings.warn(
                "I was asked to start an experiment while a current experiment is already running!"
                "I will have to stop that first! Sorry for that :)"
            )
            # Stop running experiment and wait for the spectrometer to be ready
            await self.stop_experiment()
            await self.wait_until_idle()

        start_xp_nodeid = self.opcua.get_node(self.START_EXPERIMENT).nodeid
        method_parent = self.opcua.get_node(self.METHODS)
        try:
            # Collect_bg does not seem to work in automation, set to false and do not expose in start_experiment()!
            collect_bg = False
            await method_parent.call_method(start_xp_nodeid, name, template, collect_bg)
        except ua.uaerrors.Bad as e:
            raise DeviceError(
                "The experiment could not be started!\n"
                "Check iCIR status and close any open experiment."
            ) from e
        logger.info(f"FlowIR experiment {name} started with template {template}!")

    async def stop_experiment(self):
        """Stops the experiment currently running (it does not imply instrument is then idle, wait for scan end)"""
        method_parent = self.opcua.get_node(self.METHODS)
        stop_nodeid = self.opcua.get_node(self.STOP_EXPERIMENT).nodeid
        await method_parent.call_method(stop_nodeid)

    async def wait_until_idle(self):
        """Waits until no experiment is running."""
        while await self.is_running():
            await asyncio.sleep(0.2)

    def get_router(self):
        """Creates an APIRouter for this HuberChiller instance."""
        from fastapi import APIRouter

        router = APIRouter()
        router.add_api_route("/is-connected", self.is_iCIR_connected, methods=["GET"])
        router.add_api_route("/is-running", self.is_running, methods=["GET"])
        router.add_api_route("/probe/info", self.is_iCIR_connected, methods=["GET"])
        router.add_api_route("/probe/status", self.is_iCIR_connected, methods=["GET"])
        router.add_api_route(
            "/sample/last-acquisition-time", self.last_sample_time, methods=["GET"]
        )
        router.add_api_route(
            "/sample/spectrum/last-treated", self.last_spectrum_treated, methods=["GET"]
        )
        router.add_api_route(
            "/sample/spectrum/last-raw", self.last_spectrum_raw, methods=["GET"]
        )
        router.add_api_route(
            "/sample/spectrum/last-background",
            self.last_spectrum_background,
            methods=["GET"],
        )
        router.add_api_route(
            "/experiment/start", self.start_experiment, methods=["PUT"]
        )
        router.add_api_route("/experiment/stop", self.stop_experiment, methods=["GET"])

        return router


if __name__ == "__main__":
    ...
    # async def main():
    #     opcua_client = Client(
    #         url=FlowIR.iC_OPCUA_DEFAULT_SERVER_ADDRESS.replace("localhost", "BSMC-YMEF002121")
    #     )
    #
    #     async with FlowIR(opcua_client) as ir_spectrometer:
    #         await ir_spectrometer.check_version()
    #
    #         if await ir_spectrometer.is_iCIR_connected():
    #             print("FlowIR connected!")
    #         else:
    #             raise ConnectionError("FlowIR not connected :(")
    #
    #         template_name = "15_sec_integration.iCIRTemplate"
    #         await ir_spectrometer.start_experiment(
    #             name="reaction_monitoring", template=template_name
    #         )
    #
    #         spectrum = await ir_spectrometer.last_spectrum_treated()
    #         while len(spectrum.intensity) == 0:
    #             spectrum = await ir_spectrometer.last_spectrum_treated()
    #
    #         for x in range(3):
    #             spectra_count = await ir_spectrometer.sample_count()
    #
    #             while await ir_spectrometer.sample_count() == spectra_count:
    #                 await asyncio.sleep(1)
    #
    #             print("New spectrum!")
    #             spectrum = await ir_spectrometer.last_spectrum_treated()
    #             print(spectrum)
    #
    #         await ir_spectrometer.stop_experiment()
    #
    # asyncio.run(main())



from math import pi
from warnings import warn

from flowchem.units import flowchem_ureg


class Tube:
    """
    A tube.

    Arguments:
    - `length`: The length of the tube as a str.
    - `ID`: The inner diameter of the tube as a str.
    - `OD`: The outer diameter of the tube as a str.
    - `material`: The material of the tube.

    Attributes:
    - `ID`: The inner diameter of the tube, converted to a `pint.Quantity`.
    - `length`: The length of the tube, converted to a `pint.Quantity`.
    - `material`: The material of the tube.
    - `OD`: The outer diameter of the tube, converted to a `pint.Quantity`.
    - `volume`: The volume of the tube, as determined from the length and inner diameter, converted to a `pint.Quantity`.

    Raises:
        - ValueError: When the outer diameter is less than the inner diameter of the tube.
    """

    def __init__(self, length: str, ID: str, OD: str, material: str):
        """
        See the `Tube` attributes for a description of the arguments.

        ::: tip Note
        The arguments to __init__ are `str`s, not `pint.Quantity`s.
        :::
        """
        self.length = flowchem_ureg.parse_expression(length)
        self.ID = flowchem_ureg.parse_expression(ID)
        self.OD = flowchem_ureg.parse_expression(OD)

        # check to make sure units are valid
        for measurement in [self.length, self.ID, self.OD]:
            if measurement.dimensionality != flowchem_ureg.mm.dimensionality:
                raise ValueError(
                    f"{measurement.units} is an invalid unit of measurement for length."
                )

        # ensure diameters are valid
        if self.OD <= self.ID:
            raise ValueError(
                f"Outer diameter {OD} must be greater than inner diameter {ID}"
            )
        if self.length < self.OD or self.length < self.ID:
            warn(
                f"Tube length ({self.length}) is less than diameter."
                " Make sure that this is not in error."
            )

        self.material = material
        self.volume = pi * ((self.ID / 2) ** 2) * self.length

    def __repr__(self):
        return f"Tube of length {self.length}, ID {self.ID}, OD {self.OD}"



""" Use Phidgets to control lab devices. So far, only 4..20mA interface for Swagelock Pressure-sensor """
import time
import warnings
from loguru import logger

try:
    from Phidget22.Devices.CurrentInput import CurrentInput, PowerSupply
    from Phidget22.Devices.Log import Log, LogLevel
    from Phidget22.PhidgetException import PhidgetException
except ImportError:
    HAS_PHIDGET = False
else:
    try:
        Log.enable(LogLevel.PHIDGET_LOG_INFO, "phidget.log")
    except (OSError, FileNotFoundError):
        warnings.warn(
            "Phidget22 package installed but Phidget library not found!\n"
            "Get it from https://www.phidgets.com/docs/Operating_System_Support"
        )
        HAS_PHIDGET = False
    else:
        HAS_PHIDGET = True

from flowchem.units import flowchem_ureg
from flowchem.exceptions import DeviceError, InvalidConfiguration


class PressureSensor:
    """Use a Phidget current input to translate a Swagelock 4..20mA signal to the corresponding pressure value"""

    def __init__(
        self,
        sensor_min: str = "0 bar",
        sensor_max: str = "10 bar",
        vint_serial_number: int = None,
        vint_channel: int = None,
        phidget_is_remote: bool = False,
    ):
        if not HAS_PHIDGET:
            raise InvalidConfiguration(
                "Phidget unusable: library or package not installed."
            )

        # Sensor range
        self._minP = flowchem_ureg(sensor_min)
        self._maxP = flowchem_ureg(sensor_max)
        # current meter
        self.phidget = CurrentInput()

        # Ensure connection with the right sensor (ideally these are from graph)
        if vint_serial_number:
            self.phidget.setDeviceSerialNumber(vint_serial_number)
        if vint_channel:
            self.phidget.setChannel(vint_channel)

        # Fancy remote sensors?
        if phidget_is_remote:
            from Phidget22.Net import Net
            from Phidget22.PhidgetServerType import PhidgetServerType

            Net.enableServerDiscovery(PhidgetServerType.PHIDGETSERVER_DEVICEREMOTE)
            self.phidget.setIsRemote(True)

        try:
            self.phidget.openWaitForAttachment(1000)
            logger.debug("Pressure sensor connected!")
        except PhidgetException as e:
            raise DeviceError("Cannot connect to sensor! Check settings...") from e

        # Set power supply to 24V
        self.phidget.setPowerSupply(PowerSupply.POWER_SUPPLY_24V)
        self.phidget.setDataInterval(200)  # 200ms

    def __del__(self):
        self.phidget.close()

    def get_router(self):
        """Creates an APIRouter for this object."""
        from fastapi import APIRouter

        router = APIRouter()
        router.add_api_route("/attached", self.is_attached, methods=["GET"])
        router.add_api_route("/pressure", self.read_pressure, methods=["GET"])

        return router

    def is_attached(self) -> bool:
        """Whether the device is connected"""
        return bool(self.phidget.getAttached())

    def _current_to_pressure(self, current_in_ampere: float) -> str:
        """Converts current reading into pressure value"""
        ma = current_in_ampere * 1000
        # minP..maxP is 4..20mA
        pressure_reading = self._minP + ((ma - 4) / 16) * (self._maxP - self._minP)
        logger.debug(f"Read pressure {pressure_reading} barg!")
        return str(pressure_reading * flowchem_ureg.bar)

    def read_pressure(self) -> str:
        """
        Read pressure from sensor, in bar.

        This is the main class method, and it never fails, but rather return None. Why?

        Well, initialization exception are fair play, upon startup any misconfiguration
        should be addressed.
        However, during experiment execution temporarily unavailability of a datasource
        should not be critical when the sensor is readonly.
        If the P-sensor is safety-critical than an event handler should be attached to it.
        If not we can live with it, returning None and letting the caller decide what
        to do with that.
        """
        try:
            current = self.phidget.getCurrent()
            logger.debug(f"Current pressure: {current}")
        except PhidgetException:
            warnings.warn("Cannot read pressure!")
            return ""
        else:
            return self._current_to_pressure(current)


if __name__ == "__main__":
    test = PressureSensor(
        sensor_min="0 bar", sensor_max="25 bar", vint_serial_number=627768, vint_channel=0
    )
    while True:
        print(test.read_pressure())
        time.sleep(1)



from typing import Optional

from loguru import logger

from flowchem.components.stdlib import Pump


class DummyPump(Pump):
    """
    A fake pumping device whose primary feature is that it moves fluid, used internally for testing.

    ::: warning
    Users should not instantiate a `DummyPump` for use in a `Protocol` because it is not an actual laboratory instrument.
    :::

    Arguments:
    - `name`: The name of the pump.

    Attributes:
    - `name` (`str`): The name of the pump.
    - `rate` (`pint.Quantity`): The flow rate of the pump. Must be of the dimensionality of volume/time.
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

    async def _update(self):
        logger.trace(f"Set {self} rate to {self.rate}")



from __future__ import annotations

import asyncio
import time
import traceback
from collections import namedtuple
from contextlib import AsyncExitStack
from copy import deepcopy
from time import asctime, localtime
from typing import TYPE_CHECKING, Dict, Iterable, List, Union

from loguru import logger

from flowchem.components.stdlib import ActiveComponent, Sensor
from flowchem.exceptions import ProtocolCancelled

if TYPE_CHECKING:
    from flowchem import Experiment


Datapoint = namedtuple("Datapoint", ["data", "timestamp", "experiment_elapsed_time"])


async def handle_exception(tasks_to_cancel):
    logger.error("Protocol execution is stopping NOW!")
    for task in tasks_to_cancel:
        task.cancel()
    await asyncio.sleep(5)


async def main(experiment: "Experiment", dry_run: Union[bool, int], strict: bool):
    """
    The function that actually does the execution of the protocol.

    Arguments:
    - `experiment`: The experiment to execute.
    - `dry_run`: Whether to simulate the experiment or actually perform it. If an integer greater than zero, the dry run will execute at that many times speed.
    - `strict`: Whether to stop execution upon any errors.
    """

    logger.info(f"Using Flowchem ⚗️👩‍👨🧪")
    logger.info("Performing final launch status check...")

    # Run protocol
    try:
        # To programmatically enter many context manager (one per component) AsyncExitStack is used
        async with AsyncExitStack() as stack:
            # Enter async context manager of each component. This initializes connections to hardware.
            if not dry_run:
                components = [
                    await stack.enter_async_context(compo)
                    for compo in experiment._compiled_protocol.keys()  # type:ignore
                ]
            else:
                components = list(experiment._compiled_protocol.keys())  # type:ignore

            tasks = []

            for component in components:
                # Find out when each component's monitoring should end
                procedures: Iterable = experiment._compiled_protocol[component]  # type:ignore
                end_times: List[float] = [p["time"] for p in procedures]
                end_time: float = max(end_times)  # we only want the last end time
                logger.trace(f"Calculated end time for {component} as {end_time}s")

                procedures_to_append = [
                    wait_and_execute_procedure(
                        procedure=procedure,
                        component=component,
                        experiment=experiment,
                        dry_run=dry_run,
                        strict=strict,
                    )
                    for procedure in experiment._compiled_protocol[component]  # type:ignore
                ]
                tasks.extend(procedures_to_append)
                logger.trace(f"Task list generated for {component}.")

                # for sensors, add the monitor task
                if isinstance(component, Sensor):
                    logger.trace(f"Creating sensor monitoring task for {component}")
                    tasks.append(_monitor(component, experiment, bool(dry_run), strict))
                logger.debug(f"{component} is GO!")

            logger.debug("All components are GO!")

            # Add a task to monitor the stop button
            tasks.append(check_if_cancelled(experiment))
            tasks.append(pause_handler(experiment, end_time, components))
            tasks.append(end_loop(experiment))
            logger.debug("All tasks are GO!")

            # Add a reminder about FF
            if type(dry_run) == int:
                logger.info(f"Simulating at {dry_run}x speed...")

            # begin the experiment
            logger.info("All checks passed. Experiment is GO!")
            experiment.is_executing = True
            experiment.start_time = time.time()

            # convert to local time for the start message
            _local_time = asctime(localtime(experiment.start_time))
            start_msg = f"{experiment} started at {_local_time}."

            logger.success(start_msg)

            try:
                # FIXME the list tasks actually contains coroutines, not tasks. A rename would be nice.
                task_list = [asyncio.create_task(coro) for coro in tasks]
                await asyncio.gather(*task_list)

            except ProtocolCancelled:
                logger.error("Stop button pressed.")
                await handle_exception(tasks)
                logger.critical(f"{experiment} finished by STOP button.")

            except (RuntimeError, Exception) as e:
                logger.error(f"Got {repr(e)}. Full traceback is logged at trace level.")
                await handle_exception(tasks)
                logger.critical(f"{experiment} finished by exception.")

            else:
                logger.success(f"{experiment} finished successfully.")

            finally:
                # when this code block is reached, the tasks will have either all completed or
                # an exception has occurred.
                experiment.end_time = time.time()

                # when this code block is reached, the tasks will have completed or have been cancelled.
                _local_time = asctime(localtime(experiment.end_time))
                end_msg = f"{experiment} completed at {_local_time}."

                # Stop all of the sensors and exit the read loops
                logger.debug("Resetting all components")

                # reset object
                for component in list(experiment._compiled_protocol.keys()):  # type:ignore
                    # reset object
                    logger.debug(f"Resetting {component} to base state")
                    component._update_from_params(component._base_state)

                await asyncio.sleep(1)

                # we only reach this line if things went well
                logger.info(end_msg)
    finally:

        # set some protocol metadata
        experiment.was_executed = True  # type:ignore
        # after E.was_executed=True, we THEN log that we're cleaning up so it's shown
        # in the cleanup category, not with a time in EET
        logger.info("Experimentation is over. Cleaning up...")
        experiment.is_executing = False  # type:ignore

        if experiment._bound_logger is not None:  # type:ignore
            logger.trace("Deactivating logging to Jupyter notebook widget...")
            logger.remove(experiment._bound_logger)  # type:ignore


async def wait_and_execute_procedure(
    procedure,
    component: ActiveComponent,
    experiment: "Experiment",
    dry_run: Union[bool, int],
    strict: bool,
):

    # wait for the right moment
    params = procedure["params"]
    await wait(procedure["time"], experiment, f"Set {component} to {params}")

    # NOTE: this doesn't actually call the _update() method
    component._update_from_params(params)
    logger.trace(f"{component} object state updated to reflect new params.")

    if dry_run:
        logger.info(f"Simulating: {params} on {component} at {procedure['time']}s")
    else:
        logger.info(f"Executing: {params} on {component} at {procedure['time']}s")
        try:
            await component._update()  # NOTE: This does!
        except Exception as e:
            level = "ERROR" if strict else "WARNING"
            logger.log(level, f"Failed to update {component}!")
            logger.trace(traceback.format_exc())
            if strict:
                raise RuntimeError(str(e))

    record = {
        "timestamp": time.time(),
        "params": params,
        "type": "executed_procedure" if not dry_run else "simulated_procedure",
        "component": component,
    }
    record["experiment_elapsed_time"] = record["timestamp"] - experiment.start_time

    experiment.executed_procedures.append(record)


async def _monitor(
    sensor: Sensor, experiment: "Experiment", dry_run: bool, strict: bool
):
    logger.debug(f"Started monitoring {sensor.name}")
    try:
        async for result in sensor._monitor(dry_run=dry_run, experiment=experiment):
            await experiment._update(
                device=sensor.name,
                datapoint=Datapoint(
                    data=result["data"],
                    timestamp=result["timestamp"],
                    experiment_elapsed_time=result["timestamp"] - experiment.start_time,
                ),
            )
        logger.debug(f"Stopped monitoring {sensor}")
    except Exception as e:
        logger.log("ERROR" if strict else "WARNING", f"Failed to read {sensor}!")
        logger.trace(traceback.format_exc())
        if strict:
            raise RuntimeError(str(e))


async def end_loop(experiment: "Experiment"):
    await wait(experiment.protocol._inferred_duration, experiment, "End loop")  # type:ignore
    experiment._end_loop = True  # type:ignore


async def check_if_cancelled(experiment: "Experiment") -> None:
    while not experiment._end_loop:  # type:ignore
        if experiment.cancelled:  # type:ignore
            raise ProtocolCancelled("protocol cancelled")
        await asyncio.sleep(0)


async def pause_handler(
    experiment: "Experiment", end_time: float, components: List[ActiveComponent]
) -> None:
    was_paused = False
    states: Dict[ActiveComponent, dict] = {}
    # this is either the planned duration of the experiment or cancellation
    while not experiment._end_loop:  # type:ignore

        # we need to pause
        if experiment.paused and not was_paused:
            was_paused = True
            for component in components:
                logger.debug(f"Pausing {component}.")
                states[component] = deepcopy(component.__dict__)
                component._update_from_params(component._base_state)
                await component._update()
            logger.debug("All components set to base states.")
            logger.trace(f"Saved states are {states}.")

        # we are paused but the button was hit, so we need to resume
        elif not experiment.paused and was_paused:
            logger.trace(f"Previous states: {states}")
            for component in components:
                for k, v in states[component].items():
                    setattr(component, k, v)
                await component._update()
                logger.debug(f"Reset {component} to {states[component]}.")
            was_paused = False
            states = {}
            logger.debug("All components reset to state before pause.")

        await asyncio.sleep(0)


async def wait(duration: float, experiment: "Experiment", name: str):
    """A pause-aware version of asyncio.sleep"""
    if type(experiment.dry_run) == int:
        duration /= experiment.dry_run
    await asyncio.sleep(duration)
    logger.trace(f"<{name}> Just woke up from {duration}s nap")

    while True:
        # if, at the end of sleeping, the experiment is paused, wait for it to resume
        while experiment.paused:
            await asyncio.sleep(0)
        assert not experiment.paused

        # figure out how long we've been paused for
        eet_offset = experiment._total_paused_duration
        # and where in the experimental plan we are
        assert isinstance(experiment.start_time, float)  # make the type checker happy
        eet = time.time() - experiment.start_time - eet_offset

        # do the logging thing
        logger.trace(f"Expected End Time is {eet}")
        logger.trace(f"<{name}> was supposed to execute after {duration}s")

        if (duration - eet) > 0:
            logger.trace(f"Waiting {duration - eet} more seconds")
            await asyncio.sleep(duration - eet)
        else:
            logger.trace(f"It's go time for <{name}>!")
            break



""" Router for a Spinsolve object """
from fastapi import APIRouter
from flowchem import Spinsolve


def spinsolve_get_router(device: Spinsolve) -> APIRouter:
    """Adds an APIRouter on top of an existing Spinsolve object"""
    router = APIRouter()

    @router.get("/solvent")
    async def get_solvent():
        """

        Returns:

        """
        return device.solvent

    @router.put("/solvent/{solvent_name}")
    async def set_solvent(solvent_name: str):
        """

        Args:
            solvent_name:
        """
        device.solvent = solvent_name

    @router.get("/sample-name")
    async def get_sample():
        """

        Returns:

        """
        return device.sample

    @router.put("/sample-name/{value}")
    async def set_sample(value: str):
        """

        Args:
            value:
        """
        device.sample = value

    return router



"""
This module is used to control Harvard Apparatus Elite 11 syringe pump via the 11 protocol.
"""

from __future__ import annotations
from typing import Set, Optional, List, Tuple
from loguru import logger

import asyncio
import warnings
from dataclasses import dataclass
from enum import Enum
from time import sleep

import aioserial
from pydantic import BaseModel

from flowchem.exceptions import InvalidConfiguration, DeviceError
from flowchem.units import flowchem_ureg
from flowchem.components.stdlib import Pump


class PumpInfo(BaseModel):
    """ Detailed pump info. """

    pump_type: str
    pump_description: str
    infuse_only: bool

    @classmethod
    def parse_pumpstring(cls, metrics_text: List[str]):
        """Parse pump response string into model."""
        pump_type, pump_description, infuse_only = "", "", True
        for line in metrics_text:
            if line.startswith("Pump type  "):
                pump_type = line[9:].strip()
            elif line.startswith("Pump type string"):
                pump_description = line[16:].strip()
            elif line.startswith("Direction"):
                if "withdraw" in line:
                    infuse_only = False
                else:
                    infuse_only = True
        return cls(
            pump_type=pump_type,
            pump_description=pump_description,
            infuse_only=infuse_only,
        )


@dataclass
class Protocol11Command:
    """Class representing a pump command and its expected reply"""

    command_string: str
    target_pump_address: int
    command_argument: str

    def compile(self) -> str:
        """
        Create actual command byte by prepending pump address to command.
        """
        assert 0 <= self.target_pump_address < 99
        return (
            str(self.target_pump_address)
            + self.command_string
            + " "
            + self.command_argument
            + "\r\n"
        )


class PumpStatus(Enum):
    """Possible pump statuses, as defined by the reply prompt."""

    IDLE = ":"
    INFUSING = ">"
    WITHDRAWING = "<"
    TARGET_REACHED = "T"
    STALLED = "*"


class HarvardApparatusPumpIO:
    """Setup with serial parameters, low level IO"""

    DEFAULT_CONFIG = {"timeout": 0.1, "baudrate": 115200}

    # noinspection PyPep8
    def __init__(self, port: str, **kwargs):
        # Merge default settings, including serial, with provided ones.
        configuration = dict(HarvardApparatusPumpIO.DEFAULT_CONFIG, **kwargs)

        try:
            self._serial = aioserial.AioSerial(port, **configuration)
        except aioserial.SerialException as e:
            raise InvalidConfiguration(
                f"Cannot connect to the Pump on the port <{port}>"
            ) from e

    async def _write(self, command: Protocol11Command):
        """Writes a command to the pump"""
        command_msg = command.compile()
        try:
            await self._serial.write_async(command_msg.encode("ascii"))
        except aioserial.SerialException as e:
            raise InvalidConfiguration from e
        logger.debug(f"Sent {repr(command_msg)}!")

    async def _read_reply(self) -> List[str]:
        """Reads the pump reply from serial communication"""
        reply_string = []

        for line in await self._serial.readlines_async():
            reply_string.append(line.decode("ascii").strip())
            logger.debug(f"Received {repr(line)}!")

        # First line is usually empty, but some prompts such as T* actually leak into this line sometimes.
        reply_string.pop(0)

        # remove empty strings from reply_string
        reply_string = [x for x in reply_string if x]

        return reply_string

    @staticmethod
    def parse_response_line(line: str) -> Tuple[int, PumpStatus, str]:
        """Split a received line in its components: address, prompt and reply body"""
        assert len(line) >= 3
        pump_address = int(line[0:2])
        status = PumpStatus(line[2:3])

        # Target reached is the only two-character status
        if status is PumpStatus.TARGET_REACHED:
            return pump_address, status, line[4:]
        else:
            return pump_address, status, line[3:]

    @staticmethod
    def parse_response(
        response: List[str],
    ) -> Tuple[List[int], List[PumpStatus], List[str]]:
        """Aggregates address prompt and reply body from all the reply lines and return them."""
        parsed_lines = list(map(HarvardApparatusPumpIO.parse_response_line, response))
        # noinspection PyTypeChecker
        return zip(*parsed_lines)  # type: ignore

    @staticmethod
    def check_for_errors(last_response_line, command_sent):
        """Further response parsing, checks for error messages"""
        if "Command error" in last_response_line:
            raise DeviceError(
                f"The command {command_sent} is invalid for pump {command_sent.target_pump_address}!"
                f"[Reply: {last_response_line}]"
            )
        elif "Unknown command" in last_response_line:
            raise DeviceError(
                f"The command {command_sent} is unknown to pump {command_sent.target_pump_address}!"
                f"[Maybe a withdraw command has been used with an infuse only pump?]"
                f"[Reply: {last_response_line}]"
            )
        elif "Argument error" in last_response_line:
            raise DeviceError(
                f"The command {command_sent} to pump {command_sent.target_pump_address} has an "
                f"invalid argument [Reply: {last_response_line}]"
            )
        elif "Out of range" in last_response_line:
            raise DeviceError(
                f"The command {command_sent} to pump {command_sent.target_pump_address} has an "
                f"argument out of range! [Reply: {last_response_line}]"
            )

    def reset_buffer(self):
        """Reset input buffer before reading from serial. In theory not necessary if all replies are consumed..."""
        try:
            self._serial.reset_input_buffer()
        except aioserial.PortNotOpenError as e:
            raise InvalidConfiguration from e

    async def write_and_read_reply(
        self, command: Protocol11Command, return_parsed: bool = True
    ) -> List[str]:
        """Main PumpIO method. Sends a command to the pump, read the replies and returns it, optionally parsed.

        If unparsed reply is a List[str] with raw replies.
        If parsed reply is a List[str] w/ reply body (address and prompt removed from each line)"""
        self.reset_buffer()
        await self._write(command)
        response = await self._read_reply()

        if not response:
            raise InvalidConfiguration(
                f"No response received from pump, check pump address! "
                f"(Currently set to {command.target_pump_address})"
            )

        # Parse reply
        (
            pump_address,
            return_status,
            parsed_response,
        ) = HarvardApparatusPumpIO.parse_response(response)

        # Ensures that all the replies came from the target pump (this should always be the case)
        assert all(address == command.target_pump_address for address in pump_address)

        # Ensure no stall is present (this might happen, so let's raise an Exception w/ diagnostic text)
        if PumpStatus.STALLED in return_status:
            raise DeviceError("Pump stalled! Press display on pump to clear error :(")

        HarvardApparatusPumpIO.check_for_errors(
            last_response_line=response[-1], command_sent=command
        )

        return parsed_response if return_parsed else response

    @property
    def name(self) -> Optional[str]:
        """This is used to provide a nice-looking default name to pumps based on their serial connection."""
        try:
            return self._serial.name
        except AttributeError:
            return None

    def autodetermine_address(self) -> int:
        self._serial.write("\r\n".encode("ascii"))
        self._serial.readline()
        prompt = self._serial.readline()
        address = 0 if prompt[0:2] == b":" else int(prompt[0:2])
        logger.debug(f"Address autodetected as {address}")
        return address


# noinspection SpellCheckingInspection
class Elite11Commands:

    """Holds the commands and arguments. Nota bene: Pump needs to be in Quick Start mode, which can be achieved from
    the display interface"""

    # collected commands
    # Methods can be programmed onto the pump and their execution remotely triggered.
    # No support is provided to such feature as "explicit is better than implicit", i.e. the same result can be obtained
    # with a sequence of Elite11Commands, with the advantage of ensuring code reproducibility (i.e. no specific
    # configuration is needed on the pump side)
    #
    # Other methods not included: dim display, usb echo, footswitch, poll, version (verbose ver), input,
    #                             output (if pin state high or low) and time commands

    EMPTY_MESSAGE = " "
    VERSION = "VER"

    # RUN commands (no parameters, start movement in same direction/reverse direction/infuse/withdraw respectively)
    RUN = "run"
    REVERSE_RUN = "rrun"
    INFUSE = "irun"
    WITHDRAW = "wrun"

    # STOP movement
    STOP = "stp"

    # Max applied force (in percent)
    FORCE = "FORCE"

    # Syringe diameter
    DIAMETER = "diameter"

    METRICS = "metrics"
    CURRENT_MOVING_RATE = "crate"

    # RAMP Ramping commands (infuse or withdraw)
    # setter: iramp [{start rate} {start units} {end rate} {end units} {ramp time in seconds}]
    INFUSE_RAMP = "iramp"
    GET_WITHDRAW_RAMP = "wramp"

    # RATE
    # returns or set rate irate [max | min | lim | {rate} {rate units}]
    INFUSE_RATE = "irate"
    INFUSE_RATE_LIMITS = "irate lim"
    WITHDRAW_RATE = "wrate"
    WITHDRAW_RATE_LIMITS = "wrate lim"

    # VOLUME
    SYRINGE_VOLUME = "svolume"
    INFUSED_VOLUME = "ivolume"
    WITHDRAWN_VOLUME = "wvolume"
    TARGET_VOLUME = "tvolume"

    # CLEAR VOLUME
    CLEAR_INFUSED_VOLUME = "civolume"
    CLEAR_WITHDRAWN_VOLUME = "cwvolume"
    CLEAR_INFUSED_WITHDRAWN_VOLUME = "cvolume"
    CLEAR_TARGET_VOLUME = "ctvolume"


# noinspection PyProtectedMember
class Elite11InfuseOnly(Pump):
    """
    Controls Harvard Apparatus Elite11 syringe pumps.

    The same protocol (Protocol11) can be used on other HA pumps, but is untested.
    Several pumps can be daisy chained on the same serial connection, if so address 0 must be the first one.
    Read the manufacturer manual for more details.
    """

    # This class variable is used for daisy chains (i.e. multiple pumps on the same serial connection). Details below.
    _io_instances: Set[HarvardApparatusPumpIO] = set()
    # The mutable object (a set) as class variable creates a shared state across all the instances.
    # When several pumps are daisy chained on the same serial port, they need to all access the same Serial object,
    # because access to the serial port is exclusive by definition (also locking there ensure thread safe operations).
    # FYI it is a borg idiom https://www.oreilly.com/library/view/python-cookbook/0596001673/ch05s23.html

    metadata = {
        "author": [
            {
                "first_name": "Jakob",
                "last_name": "Wolf",
                "email": "jakob.wolf@mpikg.mpg.de",
                "institution": "Max Planck Institute of Colloids and Interfaces",
                "github_username": "JB-Wolf",
            },
            {
                "first_name": "Dario",
                "last_name": "Cambie",
                "email": "dario.cambie@mpikg.mpg.de",
                "institution": "Max Planck Institute of Colloids and Interfaces",
                "github_username": "dcambie",
            },
        ],
        "stability": "beta",
        "supported": True,
    }

    def __init__(self, pump_io: HarvardApparatusPumpIO, diameter: str, syringe_volume: str,
                 address: Optional[int] = None, name: Optional[str] = None):
        """Query model and version number of firmware to check pump is
        OK. Responds with a load of stuff, but the last three characters
        are the prompt XXY, where XX is the address and Y is pump status.
        The status can be one of the three: [":", ">" "<"] respectively
        when stopped, running forwards (pumping), or backwards (withdrawing).
        The prompt is used to confirm that the address is correct.
        This acts as a check to see that the pump is connected and working."""

        self.name = f"Pump {pump_io.name}:{address}" if name is None else name
        super().__init__(name)

        self.pump_io = pump_io
        Elite11InfuseOnly._io_instances.add(self.pump_io)  # See above for details.

        self.address: int = address if address else None  # type: ignore
        self._version = None  # Set in initialize

        # diameter and syringe volume - these will be set in initialize() - check values here though.
        if diameter is None:
            raise InvalidConfiguration(
                "Please provide the syringe diameter explicitly!\nThis prevents errors :)"
            )
        else:
            self._diameter = diameter

        if syringe_volume is None:
            raise InvalidConfiguration(
                "Please provide the syringe volume explicitly!\nThis prevents errors :)"
            )
        else:
            self._syringe_volume = syringe_volume

    @classmethod
    def from_config(
        cls,
        port: str,
        diameter: str,
        syringe_volume: str,
        address: int = None,
        name: str = None,
        **serial_kwargs,
    ):
        """Programmatic instantiation from configuration

        Many pump can be present on the same serial port with different addresses.
        This shared list of PumpIO objects allow shared state in a borg-inspired way, avoiding singletons
        This is only relevant to programmatic instantiation, i.e. when from_config() is called per each pump from a
        config file, as it is the case in the HTTP server.
        Pump_IO() manually instantiated are not accounted for.
        """
        pumpio = None
        for obj in Elite11InfuseOnly._io_instances:
            if obj._serial.port == port:
                pumpio = obj
                break

        # If not existing serial object are available for the port provided, create a new one
        if pumpio is None:
            pumpio = HarvardApparatusPumpIO(port, **serial_kwargs)

        return cls(
            pumpio,
            address=address,
            name=name,
            diameter=diameter,
            syringe_volume=syringe_volume,
        )

    async def initialize(self):
        """Ensure a valid connection with the pump has been established and sets parameters."""
        # Autodetect address if none provided
        if self.address is None:
            self.address = self.pump_io.autodetermine_address()

        await self.stop()

        await self.set_syringe_diameter(self._diameter)
        await self.set_syringe_volume(self._syringe_volume)

        logger.info(
            f"Connected to pump '{self.name}' on port {self.pump_io.name}:{self.address}!"
        )

        # makes sure that a 'clean' pump is initialized.
        self._version = self._parse_version(await self.version())

        if self._version[0] >= 3:
            await self.clear_volumes()

    async def _send_command_and_read_reply(
        self, command: str, parameter="", parse=True
    ) -> str:
        """Sends a command based on its template and return the corresponding reply as str"""

        cmd = Protocol11Command(
            command_string=command,
            target_pump_address=self.address,
            command_argument=parameter,
        )
        reply = await self.pump_io.write_and_read_reply(cmd, return_parsed=parse)
        return reply[0]

    async def _send_command_and_read_reply_multiline(
        self, command: str, parameter="", parse=True
    ) -> List[str]:
        """Sends a command based on its template and return the corresponding reply as str"""

        cmd = Protocol11Command(
            command_string=command,
            target_pump_address=self.address,
            command_argument=parameter,
        )
        return await self.pump_io.write_and_read_reply(cmd, return_parsed=parse)

    async def _bound_rate_to_pump_limits(self, rate: str) -> float:
        """Bound the rate provided to pump's limit. These are function of the syringe diameter.

        NOTE: Infusion and withdraw limits are equal!"""
        # Get current pump limits (those are function of the syringe diameter)
        limits_raw = await self._send_command_and_read_reply(
            Elite11Commands.INFUSE_RATE_LIMITS
        )

        # Lower limit usually expressed in nl/min so unit-aware quantities are needed
        lower_limit, upper_limit = map(flowchem_ureg, limits_raw.split(" to "))

        # Also add units to the provided rate
        set_rate = flowchem_ureg(rate)

        # Bound rate to acceptance range
        if set_rate < lower_limit:
            warnings.warn(
                f"The requested rate {rate} is lower than the minimum possible ({lower_limit})!"
                f"Setting rate to {lower_limit} instead!"
            )
            set_rate = lower_limit

        if set_rate > upper_limit:
            warnings.warn(
                f"The requested rate {rate} is higher than the maximum possible ({upper_limit})!"
                f"Setting rate to {upper_limit} instead!"
            )
            set_rate = upper_limit

        return set_rate.to("ml/min").magnitude

    def _parse_version(self, version_text: str) -> Tuple[int, int, int]:
        """Extract semver from version string"""

        numbers = version_text.split(" ")[-1]
        version_digits = numbers.split(".")
        return int(version_digits[0]), int(version_digits[1]), int(version_digits[2])

    async def version(self) -> str:
        """Returns the current firmware version reported by the pump"""
        return await self._send_command_and_read_reply(
            Elite11Commands.VERSION
        )  # '11 ELITE I/W Single 3.0.4

    async def get_status(self) -> PumpStatus:
        """Empty message to trigger a new reply and evaluate connection and pump current status via reply prompt"""
        status = await self._send_command_and_read_reply(
            Elite11Commands.EMPTY_MESSAGE, parse=False
        )
        return PumpStatus(status[2:3])

    async def is_moving(self) -> bool:
        """Evaluate prompt for current status, i.e. moving or not"""
        prompt = await self.get_status()
        return prompt in (PumpStatus.INFUSING, PumpStatus.WITHDRAWING)

    async def is_idle(self) -> bool:
        """Returns true if idle."""
        return not await self.is_moving()

    async def get_syringe_volume(self) -> str:
        """Returns the syringe volume as str w/ units."""
        return await self._send_command_and_read_reply(
            Elite11Commands.SYRINGE_VOLUME
        )  # e.g. '100 ml'

    async def set_syringe_volume(self, volume_w_units: str = None):
        """Sets the syringe volume in ml.

        :param volume: the volume of the syringe.
        """
        volume = flowchem_ureg(volume_w_units)
        await self._send_command_and_read_reply(
            Elite11Commands.SYRINGE_VOLUME, parameter=f"{volume.m_as('ml'):.15f} m"
        )

    async def run(self):
        """Activates pump, runs in the previously set direction."""

        if await self.is_moving():
            warnings.warn("Cannot start pump: already moving!")
            return

        await self._send_command_and_read_reply(Elite11Commands.RUN)
        logger.info("Pump movement started! (direction unspecified)")

    async def infuse_run(self):
        """Activates pump, runs in infuse mode."""
        if await self.is_moving():
            warnings.warn("Cannot start pump: already moving!")
            return

        await self._send_command_and_read_reply(Elite11Commands.INFUSE)
        logger.info("Pump movement started in infuse direction!")

    async def stop(self):
        """stops pump"""
        await self._send_command_and_read_reply(Elite11Commands.STOP)
        logger.info("Pump stopped")

    async def wait_until_idle(self):
        """Wait until the pump is no more moving"""
        while await self.is_moving():
            await asyncio.sleep(0.05)

    async def get_infusion_rate(self) -> str:
        """Returns the infusion rate as str w/ units"""
        return await self._send_command_and_read_reply(
            Elite11Commands.INFUSE_RATE
        )  # e.g. '0.2 ml/min'

    async def set_infusion_rate(self, rate: str):
        """Sets the infusion rate"""
        set_rate = await self._bound_rate_to_pump_limits(rate=rate)
        await self._send_command_and_read_reply(
            Elite11Commands.INFUSE_RATE, parameter=f"{set_rate:.10f} m/m"
        )

    async def get_infused_volume(self) -> str:
        """Return infused volume as string w/ units"""
        return await self._send_command_and_read_reply(Elite11Commands.INFUSED_VOLUME)

    async def clear_infused_volume(self):
        """Reset the pump infused volume counter to 0"""
        if self._version[0] < 3:
            warnings.warn("Command not supported by pump, update firmware!")
            return
        await self._send_command_and_read_reply(Elite11Commands.CLEAR_INFUSED_VOLUME)

    async def clear_infused_withdrawn_volume(self):
        """Reset both the pump infused and withdrawn volume counters to 0"""
        self.ensure_withdraw_is_enabled()
        if self._version[0] < 3:
            warnings.warn("Command not supported by pump, update firmware!")
            return
        await self._send_command_and_read_reply(
            Elite11Commands.CLEAR_INFUSED_WITHDRAWN_VOLUME
        )
        sleep(0.1)  # FIXME check if needed

    async def clear_volumes(self):
        """Set all pump volumes to 0"""
        await self.set_target_volume(0)
        await self.clear_infused_volume()

    async def get_force(self):
        """
        Pump force, in percentage.
        Manufacturer suggested values are:
            stainless steel:    100%
            plastic syringes:   50% if volume <= 5 ml else 100%
            glass/glass:        30% if volume <= 20 ml else 50%
            glass/plastic:      30% if volume <= 250 ul, 50% if volume <= 5ml else 100%
        """
        percent = await self._send_command_and_read_reply(Elite11Commands.FORCE)
        return int(percent[:-1])

    async def set_force(self, force_percent: float):
        """Sets the pump force, see `Elite11.get_force()` for suggested values."""
        await self._send_command_and_read_reply(
            Elite11Commands.FORCE, parameter=str(int(force_percent))
        )

    async def get_syringe_diameter(self) -> str:
        """Syringe diameter in mm. This can be set in the interval 1 mm to 33 mm"""
        return await self._send_command_and_read_reply(Elite11Commands.DIAMETER)

    async def set_syringe_diameter(self, diameter_w_units: str):
        """
        Set syringe diameter. This can be set in the interval 1 mm to 33 mm
        """
        diameter = flowchem_ureg(diameter_w_units)
        if not 1 * flowchem_ureg.mm <= diameter <= 33 * flowchem_ureg.mm:
            warnings.warn(
                f"Diameter provided ({diameter}) is not valid, ignored! [Accepted range: 1-33 mm]"
            )
            return

        await self._send_command_and_read_reply(
            Elite11Commands.DIAMETER, parameter=f"{diameter.to('mm'):.4f}"
        )

    async def get_current_flowrate(self) -> str:
        """
        If pump moves, this returns the current moving rate. If not running empty string.
        :return: current moving rate
        """
        if await self.is_moving():
            return await self._send_command_and_read_reply(
                Elite11Commands.CURRENT_MOVING_RATE
            )
        else:
            warnings.warn("Pump is not moving, cannot provide moving rate!")
            return ""

    async def get_target_volume(self) -> str:
        """Returns target volume or a falsy empty string if not set."""

        target_vol = await self._send_command_and_read_reply(
            Elite11Commands.TARGET_VOLUME
        )
        if "Target volume not set" in target_vol:
            return ""
        return target_vol

    async def set_target_volume(self, volume: str):
        """
        Sets target volume in ml. If the volume is set to 0, the target is cleared.
        """
        target_volume = flowchem_ureg(volume)
        if target_volume.magnitude == 0:
            await self._send_command_and_read_reply(Elite11Commands.CLEAR_TARGET_VOLUME)
        else:
            set_vol = await self._send_command_and_read_reply(
                Elite11Commands.TARGET_VOLUME,
                parameter=f"{target_volume.m_as('ml')} m",
            )
            if "Argument error" in set_vol:
                warnings.warn(
                    f"Cannot set target volume of {target_volume} with a "
                    f"{self.get_syringe_volume()} syringe!"
                )

    async def pump_info(self) -> PumpInfo:
        """Returns many info

        e.g.
        ('Pump type          Pump 11',
        'Pump type string   11 ELITE I/W Single',
        'Display type       Sharp',
        'Steps per rev      400',
        'Gear ratio         1:1',
        'Pulley ratio       2.4:1',
        'Lead screw         24 threads per inch',
        'Microstepping      16 microsteps per step',
        'Low speed limit    27 seconds',
        'High speed limit   26 microseconds',
        'Motor polarity     Reverse',
        'Min syringe size   0.1 mm',
        'Max syringe size   33 mm',
        'Min raw force %    20%',
        'Max raw force %    80%',
        'Encoder            100 lines',
        'Direction          Infuse/withdraw',
        'Programmable       Yes',
        'Limit switches     No',
        'Command set        None', '')
        """
        parsed_multiline_response = await self._send_command_and_read_reply_multiline(
            Elite11Commands.METRICS
        )
        return PumpInfo.parse_pumpstring(parsed_multiline_response)

    def get_router(self):
        """Creates an APIRouter for this object."""
        from fastapi import APIRouter

        router = APIRouter()
        router.add_api_route(
            "/parameters/syringe-volume", self.get_syringe_volume, methods=["GET"]
        )
        router.add_api_route(
            "/parameters/syringe-volume", self.set_syringe_volume, methods=["PUT"]
        )
        router.add_api_route("/parameters/force", self.get_force, methods=["PUT"])
        router.add_api_route("/parameters/force", self.set_force, methods=["PUT"])
        router.add_api_route("/run", self.run, methods=["PUT"])
        router.add_api_route("/run/infuse", self.infuse_run, methods=["PUT"])
        router.add_api_route("/stop", self.stop, methods=["PUT"])
        router.add_api_route("/infusion-rate", self.get_infusion_rate, methods=["GET"])
        router.add_api_route("/infusion-rate", self.set_infusion_rate, methods=["PUT"])
        router.add_api_route("/info/version", self.version, methods=["GET"])
        router.add_api_route(
            "/info/status", self.get_status, methods=["GET"], response_model=PumpStatus
        )
        router.add_api_route("/info/is-moving", self.is_moving, methods=["GET"])
        router.add_api_route(
            "/info/current-flowrate", self.get_current_flowrate, methods=["GET"]
        )
        router.add_api_route(
            "/info/infused-volume", self.get_infused_volume, methods=["GET"]
        )
        router.add_api_route(
            "/info/reset-infused-volume", self.clear_infused_volume, methods=["PUT"]
        )
        router.add_api_route("/info/reset-all", self.clear_volumes, methods=["GET"])

        return router


# noinspection PyProtectedMember
class Elite11InfuseWithdraw(Elite11InfuseOnly):
    """
    Controls Harvard Apparatus Elite11 syringe pumps - INFUSE AND WITHDRAW.
    """

    def __init__(self, pump_io: HarvardApparatusPumpIO, diameter: str, syringe_volume: str,
                 address: Optional[int] = None, name: Optional[str] = None):
        """Query model and version number of firmware to check pump is
        OK. Responds with a load of stuff, but the last three characters
        are the prompt XXY, where XX is the address and Y is pump status.
        The status can be one of the three: [":", ">" "<"] respectively
        when stopped, running forwards (pumping), or backwards (withdrawing).
        The prompt is used to confirm that the address is correct.
        This acts as a check to see that the pump is connected and working."""

        super().__init__(pump_io, diameter, syringe_volume, address, name)

    async def initialize(self):
        """Ensure a valid connection with the pump has been established and sets parameters."""
        await super(Elite11InfuseWithdraw, self).initialize()

        # Additionally, ensure pump support withdrawing upon initialization
        pump_info = await self.pump_info()
        assert not pump_info.infuse_only

    async def inverse_run(self):
        """Activates pump, runs opposite to previously set direction."""
        if await self.is_moving():
            warnings.warn("Cannot start pump: already moving!")
            return

        await self._send_command_and_read_reply(Elite11Commands.REVERSE_RUN)
        logger.info("Pump movement started in reverse direction!")

    async def withdraw_run(self):
        """Activates pump, runs in withdraw mode."""
        if await self.is_moving():
            warnings.warn("Cannot start pump: already moving!")
            return

        await self._send_command_and_read_reply(Elite11Commands.WITHDRAW)

        logger.info("Pump movement started in withdraw direction!")

    async def get_withdraw_rate(self) -> str:
        """Returns the infusion rate as a string w/ units"""
        return await self._send_command_and_read_reply(Elite11Commands.WITHDRAW_RATE)

    async def set_withdraw_rate(self, rate: str):
        """Sets the infusion rate"""
        set_rate = await self._bound_rate_to_pump_limits(rate=rate)
        await self._send_command_and_read_reply(
            Elite11Commands.WITHDRAW_RATE, parameter=f"{set_rate} m/m"
        )

    async def get_withdrawn_volume(self) -> str:
        """Returns the withdrawn volume from the last clear_*_volume() command, according to the pump"""
        return await self._send_command_and_read_reply(Elite11Commands.WITHDRAWN_VOLUME)

    async def clear_withdrawn_volume(self):
        """Reset the pump withdrawn volume counter to 0"""
        await self._send_command_and_read_reply(Elite11Commands.CLEAR_WITHDRAWN_VOLUME)

    async def clear_volumes(self):
        """Set all pump volumes to 0"""
        await self.set_target_volume(0)
        await self.clear_infused_volume()
        await self.clear_withdrawn_volume()

    def get_router(self):
        router = super(Elite11InfuseWithdraw, self).get_router()
        """Creates an APIRouter for this object."""
        router.add_api_route("/run/inverse", self.inverse_run, methods=["PUT"])
        router.add_api_route("/run/withdraw", self.withdraw_run, methods=["PUT"])
        router.add_api_route("/withdraw-rate", self.get_withdraw_rate, methods=["GET"])
        router.add_api_route("/withdraw-rate", self.set_withdraw_rate, methods=["PUT"])
        router.add_api_route(
            "/info/withdrawn-volume", self.get_withdrawn_volume, methods=["GET"]
        )
        router.add_api_route(
            "/info/reset-withdrawn", self.clear_withdrawn_volume, methods=["PUT"]
        )

        return router

    async def __aenter__(self):
        await self.initialize()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.stop()

    async def _update(self):
        """Actuates flow rate changes."""
        if self.rate == 0:
            await self.stop()
        else:
            await self.set_infusion_rate(str(self.rate))
            await self.infuse_run()


if __name__ == "__main__":
    pump = Elite11InfuseOnly.from_config(port="COM4", syringe_volume="10 ml", diameter="10 mm")

    async def main():
        """Test function"""
        await pump.initialize()
        # assert await pump.get_infused_volume() == 0
        await pump.set_syringe_diameter("30 mm")
        await pump.set_infusion_rate("0.1 ml/min")
        await pump.set_target_volume("0.05 ml")
        await pump.infuse_run()
        await asyncio.sleep(2)
        await pump.pump_info()

    asyncio.run(main())



""" Autodiscover Knauer devices on network """
from typing import Tuple, Union, Text, Dict
import asyncio
import queue
import socket
import sys
import time
from loguru import logger
from threading import Thread

from getmac import getmac

Address = Tuple[str, int]


class BroadcastProtocol(asyncio.DatagramProtocol):
    """ From https://gist.github.com/yluthu/4f785d4546057b49b56c """

    def __init__(self, target: Address, response_queue: queue.Queue):
        self.target = target
        self.loop = asyncio.get_event_loop()
        self._queue = response_queue

    def connection_made(self, transport: asyncio.transports.DatagramTransport):  # type: ignore
        """Called upon connection."""
        sock = transport.get_extra_info("socket")  # type: socket.socket
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)  # sets to broadcast
        transport.sendto(b"\x00\x01\x00\xf6", self.target)

    def datagram_received(self, data: Union[bytes, Text], addr: Address):
        """Called on data received"""
        logger.trace(f"Received data from {addr}")
        self._queue.put(addr[0])


async def get_device_type(ip_address: str) -> str:
    """Returns either 'Pump', 'Valve' or 'Unknown'"""
    try:
        reader, writer = await asyncio.open_connection(host=ip_address, port=10001)
    except ConnectionError:
        return "ConnectionError"

    # Test Pump
    writer.write("HEADTYPE:?\n\r".encode())
    reply = await reader.readuntil(separator=b"\r")
    if reply.startswith(b"HEADTYPE"):
        logger.debug(f"Device {ip_address} is a Pump")
        return "Pump"

    # Test Valve
    writer.write("T:?\n\r".encode())
    reply = await reader.readuntil(separator=b"\r")
    if reply.startswith(b"VALVE"):
        logger.debug(f"Device {ip_address} is a Valve")
        return "Valve"

    return "Unknown"


def autodiscover_knauer(source_ip: str = "") -> Dict[str, str]:
    """
    Automatically find Knauer ethernet device on the network and returns the IP associated to each MAC address.
    Note that the MAC is the key here as it is the parameter used in configuration files.
    Knauer devices only support DHCP so static IPs are not an option.


    Args:
        source_ip: source IP for autodiscover (only relevant if multiple network interfaces are available!)
    Returns:
        List of tuples (IP, MAC, device_type), one per device replying to autodiscover
    """

    # Define source IP resolving local hostname.
    if not source_ip:
        hostname = socket.gethostname()
        source_ip = socket.gethostbyname(hostname)

    loop = asyncio.get_event_loop()
    device_q: queue.Queue = queue.Queue()
    coro = loop.create_datagram_endpoint(
        lambda: BroadcastProtocol(("255.255.255.255", 30718), response_queue=device_q),
        local_addr=(source_ip, 28688),
    )
    loop.run_until_complete(coro)
    thread = Thread(target=loop.run_forever)
    thread.start()
    time.sleep(2)
    loop.call_soon_threadsafe(loop.stop)  # here
    thread.join()

    device_list = []
    for _ in range(40):
        try:
            device_list.append(device_q.get_nowait())
        except queue.Empty:
            break

    device_info = dict()
    for device_ip in device_list:
        # MAC address
        mac = getmac.get_mac_address(ip=device_ip)
        device_info[mac] = device_ip
    return device_info


if __name__ == "__main__":
    # This is a bug of asyncio on Windows :|
    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    # Autodiscover devices (dict mac as index, IP as value)
    devices = autodiscover_knauer()

    for mac_address, ip in devices.items():
        # Device Type
        device_type = asyncio.run(get_device_type(ip))
        print(f"MAC: {mac_address} IP: {ip} DEVICE_TYPE: {device_type}")



""" Control module for the Vapourtec R4 heater """
import time
from loguru import logger
from typing import Tuple

import aioserial

from flowchem.exceptions import DeviceError, InvalidConfiguration
from flowchem.units import flowchem_ureg

try:
    from flowchem.components.devices.Vapourtec.commands import (
        R4Command,
        VapourtecCommand,
    )

    HAS_VAPOURTEC_COMMANDS = True
except ImportError as e:
    HAS_VAPOURTEC_COMMANDS = False
    raise PermissionError(
        "Cannot redistribute Vapourtec commands... Contact Vapourtec to get them!"
    ) from e


class R4Heater:
    """R4 reactor heater control class."""

    DEFAULT_CONFIG = {
        "timeout": 0.1,
        "baudrate": 19200,
        "parity": aioserial.PARITY_NONE,
        "stopbits": aioserial.STOPBITS_ONE,
        "bytesize": aioserial.EIGHTBITS,
    }
    """ Virtual control of the Vapourtec R4 heating module. """

    def __init__(self, **config):

        if not HAS_VAPOURTEC_COMMANDS:
            raise InvalidConfiguration(
                "R4Heater unusable: no Vapourtec Commands available.\n"
                "Contact your distributor to get the serial API documentation."
            )

        # Merge default settings, including serial, with provided ones.
        configuration = dict(R4Heater.DEFAULT_CONFIG, **config)
        try:
            self._serial = aioserial.AioSerial(**configuration)
        except aioserial.SerialException as e:
            raise InvalidConfiguration(
                f"Cannot connect to the R4Heater on the port <{config.get('port')}>"
            ) from e

    async def _write(self, command: str):
        """Writes a command to the pump"""
        cmd = command + "\r\n"
        await self._serial.write_async(cmd.encode("ascii"))
        logger.debug(f"Sent command: {repr(command)}")

    async def _read_reply(self) -> str:
        """Reads the pump reply from serial communication"""
        reply_string = await self._serial.readline_async()
        logger.debug(f"Reply received: {reply_string.decode('ascii')}")
        return reply_string.decode("ascii")

    @staticmethod
    def parse_response(response: str) -> Tuple[bool, str]:
        """Split a received line in its components: success, reply"""
        try:
            if response[0:2] != "ER":
                return True, response.rstrip()
            else:
                return False, response.rstrip()
        except KeyError:
            raise DeviceError(f"Invalid reply {response}")

    async def write_and_read_reply(self, command: R4Command) -> str:
        """Main HamiltonPumpIO method.
        Sends a command to the pump, read the replies and returns it, optionally parsed"""
        self._serial.reset_input_buffer()
        await self._write(command.compile())
        response = await self._read_reply()

        if not response:
            raise InvalidConfiguration("No response received from heating module!")

        # Parse reply
        success, parsed_response = self.parse_response(response)

        return parsed_response

    async def wait_for_target_temp(self, channel: int):
        """Waits until the target channel has reached the desired temperature and is stable"""
        t_stable = False
        failure = 0
        while not t_stable:
            try:
                ret_code = await self.write_and_read_reply(
                    VapourtecCommand.TEMP.set_argument(channel)
                )
            except InvalidConfiguration as e:
                ret_code = "N"
                failure += 1
                if failure > 3:
                    raise e
            else:
                failure = 0

            if ret_code[:1] == "S":
                logger.debug(f"Target temperature reached on channel {channel}!")
                t_stable = True
            else:
                time.sleep(1)

    async def set_temperature(
        self, channel, target_temperature: str, wait: bool = False
    ):
        """Set temperature and optionally waits for S"""
        set_command = getattr(VapourtecCommand, f"SET_CH{channel}_TEMP")

        set_temperature = flowchem_ureg(target_temperature)
        # Float not accepted, must cast to int
        await self.write_and_read_reply(
            set_command.set_argument(round(set_temperature.m_as("°C")))
        )
        # Set temperature implies channel on
        await self.write_and_read_reply(VapourtecCommand.CH_ON.set_argument(channel))

        if wait:
            await self.wait_for_target_temp(channel)

    def get_router(self):
        """Creates an APIRouter for this object."""
        from fastapi import APIRouter

        router = APIRouter()
        router.add_api_route("/temperature/set", self.set_temperature, methods=["PUT"])

        return router


if __name__ == "__main__":
    import asyncio

    heat = R4Heater(port="COM11")

    async def main():
        """test function"""
        # noinspection PyArgumentEqualDefault
        await heat.set_temperature(0, 30, wait=False)
        print("not waiting - default behaviour.")
        await heat.set_temperature(0, 30, wait=True)
        print("actually I waited")

    asyncio.run(main())



"""
This module is used to discover the serial address of any ML600 connected to the PC.
"""
import asyncio
from loguru import logger

import aioserial
import serial.tools.list_ports

from flowchem.components.devices.Hamilton.ML600 import (
    InvalidConfiguration,
    HamiltonPumpIO,
)


def ml600_finder():
    """Try to initialize an ML600 on every available COM port."""
    port_available = [comport.device for comport in serial.tools.list_ports.comports()]

    # Ports connected to an ML600-looking device
    valid_ports = set()

    for serial_port in port_available:
        try:
            print(f"Looking for pump on {serial_port}...")
            link = HamiltonPumpIO(aioserial.Serial(serial_port, timeout=0.1))
            asyncio.run(link.initialize())
            logger.info(f"{link.num_pump_connected} pump(s) found on <{serial_port}>")
            valid_ports.add(serial_port)
        except InvalidConfiguration:
            logger.debug(f"No pump found on {serial_port}")

    return valid_ports


if __name__ == "__main__":
    ml600_pumps = ml600_finder()
    if len(ml600_pumps) > 0:
        print(f"The following serial port are connected to ML600: {ml600_pumps}")
    else:
        print("No ML600 pump found")



import asyncio
from typing import Any, Dict, Optional

from loguru import logger

from flowchem.units import flowchem_ureg
from flowchem.components.stdlib import Component


class ActiveComponent(Component):
    """
    A connected, controllable component.

    All components being manipulated in a `Protocol` must be of type `ActiveComponent`.

    ::: tip
    Users should not directly instantiate an `ActiveComponent` because it is an abstract base class, not a functioning laboratory instrument.
    :::

    Arguments:
    - `name`: The name of the component.

    Attributes:
    - `name`: The name of the component.

    """

    _id_counter = 0

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)
        self._base_state: Dict[str, Any] = NotImplemented
        """
        A placeholder for the base state of the component.
        All subclasses of `ActiveComponent` must have this attribute.
        The dict must have values which can be parsed into compatible units of the object's other attributes, if applicable.
        At the end of a protocol and when not under explicit control by the user, the component will return to this state.
        """

    def _update_from_params(self, params: dict) -> None:
        """
        Updates the attributes of the object from a dict.

        Arguments:
        - `params`: A dict whose keys are the strings of attribute names and values are the new values of the attribute.
        """
        for key, value in params.items():
            if isinstance(getattr(self, key), flowchem_ureg.Quantity):
                setattr(self, key, flowchem_ureg.parse_expression(value))
            else:
                setattr(self, key, value)

    async def _update(self):
        raise NotImplementedError(f"Implement an _update() method for {repr(self)}.")

    def _validate(self, dry_run: bool) -> None:
        """
        Checks if a component's class is valid.

        Arguments:
        - `dry_run`: Whether this is a validation check for a dry run. Ignores the actual executability of the component.

        Returns:
        - Whether the component is valid or not.
        """

        logger.debug(f"Validating {self.name}...")

        # base_state method must return a dict
        if not isinstance(self._base_state, dict):
            raise ValueError("_base_state is not a dict")

        # the base_state dict must not be empty
        if not self._base_state:
            raise ValueError("_base_state dict must not be empty")

        # validate the base_state dict
        for k, v in self._base_state.items():
            if not hasattr(self, k):
                raise ValueError(
                    f"base_state sets {k} for {repr(self)} but {k} is not an attribute of {repr(self)}. "
                    f"Valid attributes are {self.__dict__}"
                )

            # dimensionality check between _base_state units and attributes
            if isinstance(self.__dict__[k], flowchem_ureg.Quantity):
                # figure out the dimensions we're comparing
                expected_dim = flowchem_ureg.parse_expression(v).dimensionality
                actual_dim = self.__dict__[k].dimensionality

                if expected_dim != actual_dim:
                    raise ValueError(
                        f"Invalid dimensionality in _base_state for {repr(self)}. "
                        f"Got {flowchem_ureg.parse_expression(v).dimensionality} for {k}, "
                        f"expected {self.__dict__[k].dimensionality}"
                    )

            # if not dimensional, do type matching
            elif not isinstance(self.__dict__[k], type(v)):
                raise ValueError(
                    f"Bad type matching for {k} in _base_state dict. "
                    f"Should be {type(self.__dict__[k])} but is {type(v)}."
                )

        # once we've checked everything, it should be good
        if not dry_run:
            self._update_from_params(self._base_state)
            logger.trace(f"Attempting to call _update() for {repr(self)}.")
            asyncio.run(self._validate_update())

        logger.debug(f"{repr(self)} is valid")

    async def _validate_update(self):
        async with self:
            res = await self._update()
            if res is not None:
                raise ValueError(f"Received return value {res} from update.")



""" Functions related to instrument reply parsing """

import warnings

from lxml import etree
from enum import Enum


class StatusNotification(Enum):
    """
    Represent the type of the status notification
    """

    STARTED = 1  # <State status="Ready"> received, starting protocol
    RUNNING = 2  # All good, <Progress> received, protocol is running
    STOPPING = 3  # Abort called, waiting for current scan end
    FINISHING = (
        4  # Upon <State status="Ready", scan acquisition over, not saving/processing
    )
    COMPLETED = 5  # Upon <Completed>, means also processing/saving data is over
    ERROR = 6  # If an error occurs
    UNKNOWN = 7


# def extract_error(xml_message: etree._Element) -> str:
#     """
#     Search for an error tag in the XML tree provided.
#     If an error is found returns its error message, empty string if no errors are present.
#     """
#     error = xml_message.find(".//Error")
#     return error.get("error") if error is not None else ""


def parse_status_notification(xml_message: etree.Element):
    """
    Parse a status notification reply.
    """
    status = xml_message.find(".//StatusNotification")

    # No status notification found
    if status is None:
        warnings.warn(
            "Parse status notification called on a message with no StatusNotification tags!"
        )
        return None

    # StatusNotification child can be <State> (w/ submsg). <Progress>, <Completed> or <Error>
    child = status[0]

    if child.tag == "State":
        return parse_state(child)

    if child.tag == "Progress":
        return StatusNotification.RUNNING, None

    if child.tag == "Completed":
        return StatusNotification.COMPLETED, None

    if child.tag == "Error":
        return StatusNotification.ERROR, None

    warnings.warn("Could not detect StatusNotification state!")
    return StatusNotification.UNKNOWN, None


def parse_state(xml_message: etree.Element):
    """Parse state message"""
    status_type = StatusNotification.UNKNOWN

    # Parse status
    status = xml_message.get("status")
    if status == "Running":
        status_type = StatusNotification.STARTED
    elif status == "Ready":
        status_type = StatusNotification.FINISHING
    elif status == "Stopping":
        status_type = StatusNotification.STOPPING
    else:
        warnings.warn(f"Unidentified notification status: {status}")

    # Full path is only available on experiment, so often this string is empty
    remote_folder = xml_message.get("dataFolder")

    if remote_folder:
        return status_type, remote_folder
    else:
        return status_type, None



""" Common iCIR code. """
import warnings
from pathlib import Path
from typing import TypedDict, List

from pydantic import BaseModel


class ProbeInfo(TypedDict):
    """Dictionary returned from iCIR with probe info."""

    spectrometer: str
    spectrometer_SN: str
    probe_SN: str
    detector: str
    apodization: str
    ip_address: str
    probe_type: str
    sampling_interval: str
    resolution: str
    scan_option: str
    gain: str


# noinspection PyPep8Naming
class iCIR_spectrometer:
    """Common code between sync and async implementations"""

    iC_OPCUA_DEFAULT_SERVER_ADDRESS = "opc.tcp://localhost:62552/iCOpcUaServer"
    _supported_versions = {"7.1.91.0"}
    SOFTWARE_VERSION = "ns=2;s=Local.iCIR.SoftwareVersion"
    CONNECTION_STATUS = "ns=2;s=Local.iCIR.ConnectionStatus"
    PROBE_DESCRIPTION = "ns=2;s=Local.iCIR.Probe1.ProbeDescription"
    PROBE_STATUS = "ns=2;s=Local.iCIR.Probe1.ProbeStatus"
    LAST_SAMPLE_TIME = "ns=2;s=Local.iCIR.Probe1.LastSampleTime"
    SAMPLE_COUNT = "ns=2;s=Local.iCIR.Probe1.SampleCount"
    SPECTRA_TREATED = "ns=2;s=Local.iCIR.Probe1.SpectraTreated"
    SPECTRA_RAW = "ns=2;s=Local.iCIR.Probe1.SpectraRaw"
    SPECTRA_BACKGROUND = "ns=2;s=Local.iCIR.Probe1.SpectraBackground"
    START_EXPERIMENT = "ns=2;s=Local.iCIR.Probe1.Methods.Start Experiment"
    STOP_EXPERIMENT = "ns=2;s=Local.iCIR.Probe1.Methods.Stop"
    METHODS = "ns=2;s=Local.iCIR.Probe1.Methods"

    def is_local(self):
        """Returns true if the server is on the same machine running the python code."""
        # noinspection PyUnresolvedReferences
        return any(
            x in self.opcua.aio_obj.server_url.netloc
            for x in ("localhost", "127.0.0.1")
        )

    @staticmethod
    def _normalize_template_name(template_name) -> str:
        """Adds .iCIRTemplate extension from string if not already present"""
        return (
            template_name
            if template_name.endswith(".iCIRTemplate")
            else template_name + ".iCIRTemplate"
        )

    @staticmethod
    def is_template_name_valid(template_name: str) -> bool:
        """
        From Mettler Toledo docs:
        You can use the Start method to create and run a new experiment in one of the iC analytical applications
        (i.e. iC IR, iC FBRM, iC Vision, iC Raman). Note that you must provide the name of an existing experiment
        template file that can be used as a basis for the new experiment.
        The template file must be located in a specific folder on the iC OPC UA Server computer.
        This is usually C:\\ProgramData\\METTLER TOLEDO\\iC OPC UA Server\\1.2\\Templates.
        """

        template_directory = Path(
            r"C:\ProgramData\METTLER TOLEDO\iC OPC UA Server\1.2\Templates"
        )
        if not template_directory.exists() or not template_directory.is_dir():
            warnings.warn("iCIR template folder not found on the local PC!")
            return False

        # Ensures the name has been provided with no extension (common mistake)
        template_name = iCIR_spectrometer._normalize_template_name(template_name)
        for existing_template in template_directory.glob("*.iCIRTemplate"):
            if existing_template.name == template_name:
                return True
        return False

    @staticmethod
    def parse_probe_info(probe_info_reply: str) -> ProbeInfo:
        """Convert the device reply into a ProbeInfo dictionary

        Example probe_info_reply reply is:
        'FlowIR; SN: 2989; Detector: DTGS; Apodization: HappGenzel; IP Address: 192.168.1.2;
        Probe: DiComp (Diamond); SN: 14570173; Interface: FlowIR™ Sensor; Sampling: 4000 to 650 cm-1;
        Resolution: 8; Scan option: AutoSelect; Gain: 232;'
        """
        fields = probe_info_reply.split(";")
        probe_info = {
            "spectrometer": fields[0],
            "spectrometer_SN": fields[1].split(": ")[1],
            "probe_SN": fields[6].split(": ")[1],
        }

        # Use aliases, i.e. translate API names (left) to dict key (right)
        translate_attributes = {
            "Detector": "detector",
            "Apodization": "apodization",
            "IP Address": "ip_address",
            "Probe": "probe_type",
            "Sampling": "sampling_interval",
            "Resolution": "resolution",
            "Scan option": "scan_option",
            "Gain": "gain",
        }
        for element in fields:
            if ":" in element:
                piece = element.split(":")
                if piece[0].strip() in translate_attributes:
                    probe_info[translate_attributes[piece[0].strip()]] = piece[
                        1
                    ].strip()

        return probe_info  # type: ignore


class IRSpectrum(BaseModel):
    """
    IR spectrum class.
    Consider rampy for advance features (baseline fit, etc)
    See e.g. https://github.com/charlesll/rampy/blob/master/examples/baseline_fit.ipynb
    """

    wavenumber: List[float]
    intensity: List[float]



from flowchem.components.stdlib import Dummy


class BrokenDummyComponent(Dummy):
    """
    A fake component, used internally for testing. Its `_update()` method always returns `True` during a dry run but always returns `False` in a real run.

    ::: danger
    Using this component during real protocol execution will result in a failure.
    :::

    Arguments:
    - `name`: The name of the component.

    Attributes:
    - `active`: Whether the component is active. This doesn't actually mean anything.
    """

    def __init__(self, name=None):
        super().__init__(name=name)

    async def _update(self) -> None:
        if self.active:
            raise RuntimeError

    def _validate(self, dry_run):
        return True



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def get_version(rel_path: str) -> str:
    """ Get __version__ without importing init """
    with open(rel_path, 'r') as f:
        for line in f.read().splitlines():
            if line.startswith("__version__"):
                # __version__ = "0.9"
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name="flowchem",
    version=get_version("flowchem/__init__.py"),
    author="Dario Cambié, Jakob Wolf",
    author_email="dario.cambie@mpikg.mpg.de, jakob.wolf@mpikg.mpg.de",
    description="Flowchem is a python library to control a variety of instruments commonly found in chemistry labs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cambiegroup/flowchem",
    packages=setuptools.find_packages(),
    install_requires=[
        "pyserial",
        "pint",
        "pandas",
        "scipy",
        "numpy",
        "asyncua",
        "phidget22",
        "getmac",
        "lmfit",
        "nmrglue",
        "aioserial",
        "jsonschema",
        "loguru",
        "networkx",
        "ipython",
        "altair",
        "aiofiles",
        "pyyaml",
        "bokeh",
        "graphviz",
        "ord-schema",
        "google-api-python-client"
    ],
    python_requires=">=3.8",
    entry_points={"console_scripts": ["flowchem=flowchem.cli:main"]},
)



""" Exceptions used in the flowchem module. """


class DeviceError(BaseException):
    """Generic DeviceError"""

    pass


class InvalidConfiguration(DeviceError):
    """The configuration provided is not valid, e.g. no connection w/ device obtained"""

    pass


class ActuationError(DeviceError):
    """The attepted move did not succeed"""

    pass


class ProtocolCancelled(Exception):
    pass



from .Harvard_Apparatus import *
from .Hamilton import *
from .Huber import *
from .Knauer import *
from .Magritek import *
from .Manson import *
from .MettlerToledo import *
from .Phidgets import *
from .Vapourtec import *
from .ViciValco import *



""" Spinsolve module """
import asyncio
import pprint as pp
import queue
import threading
import warnings
from loguru import logger
from pathlib import Path
from typing import Optional, Union

from lxml import etree
from packaging import version

from flowchem.components.devices.Magritek.msg_maker import (
    create_message,
    create_protocol_message,
    set_user_data,
    set_data_folder,
    set_attribute,
    get_request,
)
from flowchem.components.devices.Magritek.parser import (
    parse_status_notification,
    StatusNotification,
)
from flowchem.components.devices.Magritek.reader import Reader
from flowchem.components.devices.Magritek.utils import (
    run_loop,
    get_streams_for_connection,
)


class Spinsolve:
    """
    Spinsolve class, gives access to the spectrometer remote control API
    """

    default_port = 13000

    def __init__(
        self,
        io_reader: asyncio.StreamReader,
        io_writer: asyncio.StreamWriter,
        loop,
        **kwargs,
    ):
        """
        Constructor, actuates the connection upon instantiation.

        Dependency injection is used here with async stream reader/writer.
        This also simplifies testing.
        """

        # IOs
        self._io_writer = io_writer
        self._io_reader = io_reader
        # Queue needed for thread-safe operation, the reader is in a different thread
        self._replies: queue.Queue = queue.Queue()
        self._reader = Reader(self._replies, kwargs.get("xml_schema", None))

        # Event loop to run the connection listener
        self._loop = loop
        self._loop.create_task(self.connection_listener())
        threading.Thread(target=lambda: run_loop(self._loop), daemon=True).start()

        # Check if the instrument is connected
        hw_info = self.hw_request()

        # If not raise ConnectionError
        if hw_info.find(".//ConnectedToHardware").text != "true":
            raise ConnectionError(
                "The spectrometer is not connected to the control PC running Spinsolve software!"
            )

        # If connected parse and log instrument info
        self.software_version = hw_info.find(".//SpinsolveSoftware").text
        self.hardware_type = hw_info.find(".//SpinsolveType").text
        logger.debug(
            f"Connected with Spinsolve model {self.hardware_type}, SW version: {self.software_version}"
        )

        # Load available protocols
        self.protocols_option = self.request_available_protocols()

        # Set experimental variable
        self.data_folder = kwargs.get("data_folder")

        # An optional mapping between remote and local folder location can be used for remote use
        self._folder_mapper = kwargs.get("remote_to_local_mapping")
        # Ensure mapper validity before starting. This will save you time later ;)
        if self._folder_mapper is not None:
            assert self._folder_mapper(self.data_folder) is not None

        # Sets default sample, solvent value and user data
        self.sample = kwargs.get("sample_name", "FlowChem Experiment")
        self.solvent = kwargs.get("solvent", "Chloroform?")
        self.user_data = dict(control_software="flowchem")

        # Finally check version
        if version.parse(self.software_version) < version.parse("1.18.1.3062"):
            warnings.warn(
                f"Spinsolve version {self.software_version} is older than the reference (1.18.1.3062)"
            )

    @classmethod
    def from_config(cls, config):
        """Create instance from config dict"""
        # Get loop if passed or existing
        loop = config.get("loop", asyncio.get_event_loop())
        # Create stream reader/writer pair
        reader, writer = loop.run_until_complete(
            get_streams_for_connection(config.get("host"), config.get("port", "13000"))
        )
        # Drop keys
        for key in ("host", "port", "loop"):
            config.pop(key, None)
        # Instantiate with the rest of config
        return cls(reader, writer, loop, **config)

    def __del__(self):
        self._loop.stop()

    async def connection_listener(self):
        """
        Listen for replies and puts them in the queue
        """
        while True:
            try:
                chunk = await self._io_reader.read(1024)
            except asyncio.CancelledError:
                break
            self._replies.put(chunk)

    def _transmit(self, message: bytes):
        """
        Sends the message to the spectrometer
        """
        self._io_writer.write(message)

    @property
    def solvent(self) -> str:
        """Get current solvent"""
        # Send request
        self.send_message(get_request("Solvent"))

        # Get reply
        reply = self._read_reply(reply_type="GetResponse")

        # Parse and return
        return reply.find(".//Solvent").text

    @solvent.setter
    def solvent(self, solvent: str):
        """Sets the solvent"""
        self.send_message(set_attribute("Solvent", solvent))

    @property
    def sample(self) -> str:
        """Get current solvent (appears in acqu.par)"""
        # Send request
        self.send_message(get_request("Sample"))

        # Get reply
        reply = self._read_reply(reply_type="GetResponse")

        # Parse and return
        return reply.find(".//Sample").text

    @sample.setter
    def sample(self, sample: str):
        """Sets the sample name (appears in acqu.par)"""
        self.send_message(set_attribute("Sample", sample))

    @property
    def data_folder(self):
        """Get current data_folder location"""
        return self._data_folder  # No command available to directly query Spinsolve :(

    @data_folder.setter
    def data_folder(self, location: str):
        """Sets the location provided as data folder. optionally, with typeThese are included in acq.par"""
        if location is not None:
            self._data_folder = location
            self.send_message(set_data_folder(location))

    @property
    def user_data(self) -> dict:
        """Create a get user data request and parse result"""
        # Send request
        self.send_message(get_request("UserData"))

        # Get reply
        reply = self._read_reply(reply_type="GetResponse")

        # Parse and return
        return {
            data_item.get("key"): data_item.get("value")
            for data_item in reply.findall(".//Data")
        }

    @user_data.setter
    def user_data(self, data_to_be_set: dict):
        """Sets the user data proewqvided in the dict. These are included in acq.par"""
        self.send_message(set_user_data(data_to_be_set))

    def _read_reply(self, reply_type="", timeout=5):
        """Looks in the received replies for one of type reply_type"""
        # Get reply of reply_type from the reader object that holds the StreamReader
        reply = self._reader.wait_for_reply(reply_type=reply_type, timeout=timeout)
        logger.debug(f"Got a reply from spectrometer: {etree.tostring(reply)}")

        return reply

    async def _async_read_reply(self, *args):
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, self._read_reply, *args)

    def send_message(self, root: etree.Element):
        """
        Sends the tree connected XML etree.Element provided
        """
        # Turn the etree.Element provided into an ElementTree
        tree = etree.ElementTree(root)

        # Export to string with declaration
        message = etree.tostring(tree, xml_declaration=True, encoding="utf-8")

        # Transmit
        logger.debug(f"Transmitting request to spectrometer: {message}")
        self._transmit(message)

    def hw_request(self):
        """
        Sends an HW request to the spectrometer, receive the reply and returns it
        """
        self.send_message(create_message("HardwareRequest"))
        # Larger timeout than usual as this is the first request sent to the spectrometer
        # When the PC/Spinsolve is in standby, the reply to the first req is slower than usual
        return self._read_reply(reply_type="HardwareResponse", timeout=15)

    def request_available_protocols(self) -> dict:
        """
        Get a list of available protocol on the current spectrometer
        """
        # Request available protocols
        self.send_message(create_message("AvailableProtocolOptionsRequest"))
        # Get reply
        tree = self._read_reply(reply_type="AvailableProtocolOptionsResponse")

        # Parse reply and construct the dict with protocols available
        protocols = dict()
        for element in tree.findall(".//Protocol"):
            protocol_name = element.get("protocol")
            protocols[protocol_name] = {
                option.get("name"): [value.text for value in option.findall("Value")]
                for option in element.findall("Option")
            }

        return protocols

    async def run_protocol(
        self, protocol_name, protocol_options=None
    ) -> Optional[Union[str, Path]]:
        """
        Runs a protocol

        Returns true if the protocol is started correctly, false otherwise.
        """
        # All protocol names are UPPERCASE, so force upper here to avoid case issues
        protocol_name = protocol_name.upper()
        if not self.is_protocol_available(protocol_name):
            warnings.warn(
                f"The protocol requested '{protocol_name}' is not available on the spectrometer!\n"
                f"Valid options are: {pp.pformat(sorted(self.protocols_option.keys()))}"
            )
            return None

        # Validate protocol options (check values and remove invalid ones, with warning)
        valid_protocol_options = self._validate_protocol_request(
            protocol_name, protocol_options
        )

        # Start protocol
        self._reader.clear_replies()
        self.send_message(
            create_protocol_message(protocol_name, valid_protocol_options)
        )

        # Follow status notifications and finally get location of remote data
        remote_data_folder = await self.check_notifications()
        logger.info(f"Protocol over - remote data folder is {remote_data_folder}")

        # If a folder mapper is present use it to translate the location
        if self._folder_mapper:
            return self._folder_mapper(remote_data_folder)
        else:
            return remote_data_folder

    async def check_notifications(self) -> Path:
        """
        Read all the StatusNotification and returns the dataFolder
        """
        remote_folder = Path(".")
        while True:
            # Get all StatusNotification
            # status_update = self._read_reply("StatusNotification", 6000)
            status_update = await self._async_read_reply("StatusNotification", 6000)

            # Parse them
            status, folder = parse_status_notification(status_update)
            logger.debug(
                f"Status update: Status is {status} and data folder={folder}"
            )

            # When I get a response with folder, save the location!
            if folder:
                remote_folder = Path(folder)

            # If it is complete leave loop
            if status is StatusNotification.COMPLETED:
                break

            if status is StatusNotification.ERROR:
                warnings.warn(
                    "Error detected on running protocol -- aborting."
                )  # Usually device busy
                self.abort()  # Abort running experiment
                break
        return remote_folder

    def abort(self):
        """Abort current running command"""
        self.send_message(create_message("Abort"))

    def is_protocol_available(self, desired_protocol):
        """Check if the desired protocol is available on the current instrument"""
        return desired_protocol in self.protocols_option

    def _validate_protocol_request(self, protocol_name, protocol_options) -> dict:
        """Ensures the protocol names, option name and option values are valid."""
        # Valid option for protocol
        valid_options = self.protocols_option.get(protocol_name)
        if valid_options is None or protocol_options is None:
            return dict()

        # For each option, check if valid. If not, remove it, raise warning and continue
        for option_name, option_value in list(protocol_options.items()):
            if option_name not in valid_options:
                protocol_options.pop(option_name)
                warnings.warn(
                    f"Invalid option {option_name} for protocol {protocol_name} -- DROPPED!"
                )
                continue

            # Get valid option values (list of them or empty list if not a multiple choice)
            valid_values = valid_options[option_name]

            # If there is no list of valid options accept anything
            if not valid_values:
                continue
            # otherwise validate the value as well
            elif str(option_value) not in valid_values:
                protocol_options.pop(option_name)
                warnings.warn(
                    f"Invalid value {option_value} for option {option_name} in protocol {protocol_name}"
                    f" -- DROPPED!"
                )

        # Returns the dict with only valid options/value pairs
        return protocol_options

    # def shim(self, shim_type="CheckShim") -> Tuple[float, float]:
    #     """ Perform one of the standard shimming routine {CheckShim | QuickShim | PowerShim} """
    #     # Check shim type
    #     if shim_type not in self.STD_SHIM_REQUEST:
    #         warnings.warn(f"Invalid shimming protocol: {shim_type} not in {self.STD_SHIM_REQUEST}. Assumed CheckShim")
    #         shim_type = "CheckShim"
    #
    #     # Submit request
    #     self.send_message(create_message(shim_type+"Request"))
    #
    #     # Wait for reply
    #     response_tag = shim_type + "Response"
    #     wait_time = {
    #         "CheckShim": 180,
    #         "QuickShim": 600,
    #         "PowerShim": 3600,
    #     }
    #     reply = self._read_reply(reply_type=response_tag, timeout=wait_time[shim_type])
    #
    #     # Check for errors
    #     error = reply.find(f".//{response_tag}").get("error")
    #     if error:
    #         warnings.warn(f"Error occurred during shimming: {error}")
    #         return None, None
    #
    #     # Return LineWidth and BaseWidth
    #     return float(reply.find(".//LineWidth").text), float(reply.find(".//BaseWidth").text)

    def shim(self):
        """Performs a shim on sample"""
        raise NotImplementedError("Use run protocol with a shimming protocol instead!")



""" Knauer valve control. """

import warnings
from loguru import logger
from enum import Enum

from flowchem.exceptions import DeviceError
from flowchem.components.devices.Knauer.Knauer_common import KnauerEthernetDevice
from flowchem.components.stdlib.valve import Valve


class KnauerValveHeads(Enum):
    """
    Four different valve types can be used. 6port2position valve, and 6, 12, 16 multi-position valves
    """

    SIX_PORT_TWO_POSITION = "LI"
    SIX_PORT_SIX_POSITION = "6"
    TWELVE_PORT_TWELVE_POSITION = "12"
    SIXTEEN_PORT_SIXTEEN_POSITION = "16"


class KnauerValve(KnauerEthernetDevice, Valve):
    """
    Control Knauer multi position valves.

    Valve type can be 6, 12, 16
    or it can be 6 ports, two positions, which will be simply 2 (two states)
    in this case, the response for T is LI. Load and inject can be switched by sending L or I
    maybe valves should have an initial state which is set during init and updated, if no  change don't schedule command
    EN: https://www.knauer.net/Dokumente/valves/azura/manuals/v6860_azura_v_2.1s_user-manual_en.pdf
    DE: https://www.knauer.net/Dokumente/valves/azura/manuals/v6860_azura_v_2.1s_benutzerhandbuch_de.pdf
    DIP switch for valve selection
    """

    def __init__(self, ip_address=None, mac_address=None, name=None):
        super().__init__(ip_address, mac_address, name)
        self.eol = b"\r\n"
        # These are set during initialize()
        self.valve_type = None
        self._position = None

    async def initialize(self):
        """Initialize connection"""
        # Here the magic happens...
        await super().initialize()

        # Detect valve type and state
        self.valve_type = await self.get_valve_type()
        self._position = await self.get_current_position()

    @staticmethod
    def handle_errors(reply: str):
        """True if there are errors, False otherwise. Warns for errors."""

        if not reply.startswith("E"):
            return None

        if "E0" in reply:
            DeviceError(
                "The valve refused to switch.\n"
                "Replace the rotor seals of the valve or replace the motor drive unit."
            )
        elif "E1" in reply:
            DeviceError(
                "Skipped switch: motor current too high!\n"
                "Replace the rotor seals of the valve."
            )
        elif "E2" in reply:
            DeviceError(
                "Change from one valve position to the next takes too long.\n"
                "Replace the rotor seals of the valve."
            )
        elif "E3" in reply:
            DeviceError(
                "Switch position of DIP 3 and 4 are not correct.\n"
                "Correct DIP switch 3 and 4."
            )
        elif "E4" in reply:
            DeviceError(
                "Valve homing position not recognized.\n" "Readjust sensor board."
            )
        elif "E5" in reply:
            DeviceError(
                "Switch position of DIP 1 and 2 are not correct.\n"
                "Correct DIP switch 1 and 2."
            )
        elif "E6" in reply:
            DeviceError("Memory error.\n" "Power-cycle valve!")
        else:
            DeviceError("Unspecified error detected!")

    async def _transmit_and_parse_reply(self, message: str) -> str:
        """
        Sends command, receives reply and parse it.

        :param message: str with command to be sent
        :return: reply: str with reply
        """
        reply = await self._send_and_receive(message)
        self.handle_errors(reply)

        if reply == "?":
            # retry once before failing. This happens often on pos commands!
            reply = await self._send_and_receive(message)
            if reply == "?":
                warnings.warn(f"Command failed: {message}")
                logger.warning(f"Command failed: {message}")
                return ""

        return reply

    async def get_current_position(self) -> str:
        """Return current valve position."""
        return await self._transmit_and_parse_reply("P")

    async def switch_to_position(self, position: str):
        """Move valve to position."""
        position = position.upper()

        # switching necessary?
        if position == self._position:
            logger.debug("Target position == current position. No movement needed.")
            return

        # Switch to position
        await self._transmit_and_parse_reply(position)

    async def get_valve_type(self):
        """
        Gets valve type, if returned value is not supported throws an error.

        Note that this method is called during initialize(), therefore it is in line
        with the general philosophy of the module to 'fail early' upon init and avoiding
        raising exception after that.
        """
        reply = await self._transmit_and_parse_reply("T")
        assert reply.startswith("VALVE ")
        reply = reply[6:]

        try:
            headtype = KnauerValveHeads(reply)
        except ValueError as e:
            raise DeviceError(
                "The valve type returned is not recognized.\n"
                "Are you sure the address provided is correct?\n"
                "Only multi-pos 6, 12, 16 and 2-pos 6 port valves are supported!"
            ) from e

        logger.info(f"Valve connected, type: {headtype}.")
        return headtype


class Knauer6Port2PositionValve(KnauerValve):
    """KnauerValve of type SIX_PORT_TWO_POSITION"""

    async def initialize(self):
        """Ensure valve type"""
        await super().initialize()
        assert self.valve_type == KnauerValveHeads.SIX_PORT_TWO_POSITION


class Knauer6Port6PositionValve(KnauerValve):
    """KnauerValve of type SIX_PORT_SIX_POSITION"""

    async def initialize(self):
        """Ensure valve type"""
        await super().initialize()
        assert self.valve_type == KnauerValveHeads.SIX_PORT_SIX_POSITION


class Knauer12PortValve(KnauerValve):
    """KnauerValve of type TWELVE_PORT_TWELVE_POSITION"""

    async def initialize(self):
        """Ensure valve type"""
        await super().initialize()
        assert self.valve_type == KnauerValveHeads.TWELVE_PORT_TWELVE_POSITION


class Knauer16PortValve(KnauerValve):
    """KnauerValve of type SIXTEEN_PORT_SIXTEEN_POSITION"""

    async def initialize(self):
        """Ensure valve type"""
        await super().initialize()
        assert self.valve_type == KnauerValveHeads.SIXTEEN_PORT_SIXTEEN_POSITION


if __name__ == "__main__":
    # This is a bug of asyncio on Windows :|
    import sys
    import asyncio

    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    v = KnauerValve(ip_address="192.168.1.176")

    async def main(valve: KnauerValve):
        """test function"""
        await valve.initialize()
        await valve.switch_to_position("I")
        print(await valve.get_current_position())
        await valve.switch_to_position("L")
        print(await valve.get_current_position())

    asyncio.run(main(v))



from typing import Union, Optional, Mapping, MutableMapping, List, Dict, Any, Iterable

import json
import os
from copy import deepcopy
from datetime import timedelta
from math import isclose
from warnings import warn

import altair as alt
import pandas as pd
import yaml
from IPython import get_ipython
from IPython.display import Code
from loguru import logger

from flowchem.units import flowchem_ureg
from flowchem.components.stdlib import ActiveComponent, TempControl, Valve
from flowchem.core.apparatus import Apparatus
from flowchem.core.experiment import Experiment


class Protocol(object):
    """
    A set of procedures for an apparatus.

    A protocol is defined as a list of procedures, atomic steps for the individual active components of an apparatus.

    ::: tip
    The same `Apparatus` object can create multiple distinct `Protocol` objects.
    :::

    Arguments:
    - `apparatus`: The apparatus for which the protocol is being defined.
    - `name`: The name of the protocol. Defaults to "Protocol_X" where *X* is protocol count.
    - `description`: A longer description of the protocol.

    Attributes:
    - `apparatus`: The apparatus for which the protocol is being defined.
    - `description`: A longer description of the protocol.
    - `is_executing`: Whether the protocol is executing.
    - `name`: The name of the protocol. Defaults to "Protocol_X" where *X* is protocol count.
    - `procedures`: A list of the procedures for the protocol in which each procedure is a dict.
    - `was_executed`: Whether the protocol was executed.
    """

    _id_counter = 0

    def __init__(
        self,
        apparatus: Apparatus,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ):
        """See main docstring."""
        # type checking
        if not isinstance(apparatus, Apparatus):
            raise TypeError(
                f"Must pass an Apparatus object. Got {type(apparatus)}, "
                "which is not an instance of flowchem.Apparatus."
            )

        # ensure apparatus is valid
        if not apparatus._validate():
            raise ValueError("Apparaus is not valid.")

        # store the passed args
        self.apparatus = apparatus
        self.description = description

        # generate the name
        if name is not None:
            self.name = name
        else:
            self.name = "Protocol_" + str(Protocol._id_counter)
            Protocol._id_counter += 1

        # default values
        self.procedures: List[
            Dict[str, Union[float, None, ActiveComponent, Dict[str, Any]]]
        ] = []

    def __repr__(self):
        return f"<{self.__str__()}>"

    def __str__(self):
        return f"Protocol {self.name} defined over {repr(self.apparatus)}"

    def _check_added_valve_mapping(self, valve: Valve, **kwargs) -> dict:
        setting = kwargs["setting"]

        if valve.mapping is None:
            raise ValueError(f"{repr(valve)} does not have a mapping.")

        # the valve itself was given
        if setting in valve.mapping:
            logger.trace(f"{setting} in {repr(valve)}'s mapping.")
            kwargs["setting"] = valve.mapping[setting]

        # the valve's name was given
        # in this case, we get the mapped valve with that name
        # we don't have to worry about duplicate names since that's checked later
        elif setting in [c.name for c in valve.mapping]:
            logger.trace(f"{setting} in {repr(valve)}'s mapping.")
            mapped_component = [c for c in valve.mapping if c.name == setting]
            kwargs["setting"] = valve.mapping[mapped_component[0]]

        # the user gave the actual port mapping number
        elif setting in valve.mapping.values() and isinstance(setting, int):
            logger.trace(f"User supplied manual setting for {valve}")
        else:
            raise ValueError(f"Invalid setting {setting} for {repr(valve)}.")

        return kwargs

    def _check_component_kwargs(self, component: ActiveComponent, **kwargs) -> None:
        """Checks that the given keyword arguments are valid for a component."""
        for kwarg, value in kwargs.items():
            # check that the component even has the attribute
            if not hasattr(component, kwarg):
                # id nor determine valid attrs for the error message
                valid_attrs = [x for x in vars(component).keys()]
                # we don't care about the name attr
                valid_attrs = [x for x in valid_attrs if x != "name"]
                # or internal ones
                valid_attrs = [x for x in valid_attrs if not x.startswith("_")]

                msg = f"Invalid attribute {kwarg} for {component}. "
                msg += f"Valid attributes are {valid_attrs}"
                raise ValueError(msg)

            # for kwargs that will be converted later, just check that the units match
            if isinstance(component.__dict__[kwarg], flowchem_ureg.Quantity):
                try:
                    value_dim = flowchem_ureg.parse_expression(value).dimensionality
                except AttributeError:
                    value_dim = type(value)
                kwarg_dim = component.__dict__[kwarg].dimensionality

                # perform the check
                if value_dim != kwarg_dim:
                    msg = f"Bad dimensionality of {kwarg} for {component}. "
                    msg += f"Expected {kwarg_dim} but got {value_dim}."
                    raise ValueError(msg)

            # if it's not a quantity, check the types
            elif not isinstance(value, type(component.__dict__[kwarg])):
                expected_type = type(component.__dict__[kwarg])
                msg = "Bad type matching. "
                msg += f"Expected '{kwarg}' to an instance of {expected_type} but got"
                msg += f"{repr(value)}, which is of type {type(value)}."
                raise ValueError(msg)

    def _add_single(
        self, component: ActiveComponent, start=None, stop=None, duration=None, **kwargs
    ) -> None:
        """Adds a single procedure to the protocol.

        See add() for full documentation.
        """

        # make sure that the component being added is part of the apparatus
        self.apparatus[component]

        # don't let users give empty procedures
        if not kwargs:
            raise RuntimeError(
                "No kwargs supplied. "
                "This will not manipulate the state of your sythesizer. "
                "Ensure your call to add() is valid."
            )

        # perform the mapping for valves
        if isinstance(component, Valve):
            kwargs = self._check_added_valve_mapping(component, **kwargs)

        # make sure the component and keywords are valid
        self._check_component_kwargs(component, **kwargs)

        if stop is not None and duration is not None:
            raise RuntimeError("Must provide one of stop and duration, not both.")

        # parse the start time if given
        if isinstance(start, timedelta):
            start = str(start.total_seconds()) + " seconds"
        elif start is None:  # default to the beginning of the protocol
            start = "0 seconds"
        start = flowchem_ureg.parse_expression(start)

        # parse duration if given
        if duration is not None:
            if isinstance(duration, timedelta):
                duration = str(duration.total_seconds()) + " seconds"
            stop = start + flowchem_ureg.parse_expression(duration)
        elif stop is not None:
            if isinstance(stop, timedelta):
                stop = str(stop.total_seconds()) + " seconds"
            if isinstance(stop, str):
                stop = flowchem_ureg.parse_expression(stop)

        if stop is not None and start > stop:
            raise ValueError("Procedure beginning is after procedure end.")

        # a little magic for temperature controllers
        if isinstance(component, TempControl):
            if kwargs.get("temp") is not None and kwargs.get("active") is None:
                kwargs["active"] = True
            elif not kwargs.get("active") and kwargs.get("temp") is None:
                kwargs["temp"] = "0 degC"
            elif kwargs["active"] and kwargs.get("temp") is None:
                raise RuntimeError(
                    f"TempControl {component} is activated but temperature "
                    "setting is not given. Specify 'temp' in your call to add()."
                )

        # add the procedure to the procedure list
        self.procedures.append(
            dict(
                start=float(start.to_base_units().magnitude)
                if start is not None
                else start,
                stop=float(stop.to_base_units().magnitude)
                if stop is not None
                else stop,
                component=component,
                params=kwargs,
            )
        )

    def add(
        self,
        component: Union[ActiveComponent, Iterable[ActiveComponent]],
        start=None,
        stop=None,
        duration=None,
        **kwargs,
    ):
        """
        Adds a procedure to the protocol.

        ::: warning
        If stop and duration are both `None`, the procedure's stop time will be inferred as the end of the protocol.
        :::

        Arguments:
        - `component_added`: The component(s) for which the procedure being added. If an interable, all components will have the same parameters.
        - `start`: The start time of the procedure relative to the start of the protocol, such as `"5 seconds"`. May also be a `datetime.timedelta`. Defaults to `"0 seconds"`, *i.e.* the beginning of the protocol.
        - `stop`: The stop time of the procedure relative to the start of the protocol, such as `"30 seconds"`. May also be a `datetime.timedelta`. May not be given if `duration` is used.
        duration: The duration of the procedure, such as "1 hour". May not be used if `stop` is used.
        - `**kwargs`: The state of the component for the procedure.

        Raises:
        - `TypeError`: A component is not of the correct type (*i.e.* a Component object)
        - `ValueError`: An error occurred when attempting to parse the kwargs.
        - `RuntimeError`: Stop time of procedure is unable to be determined or invalid component.
        """

        if isinstance(component, Iterable):
            for _component in component:
                self._add_single(
                    _component, start=start, stop=stop, duration=duration, **kwargs
                )
        else:
            self._add_single(
                component, start=start, stop=stop, duration=duration, **kwargs
            )

    @property
    def _inferred_duration(self):
        # infer the duration of the protocol
        computed_durations = sorted(
            [x["stop"] for x in self.procedures],
            key=lambda z: z if z is not None else 0,
        )
        if all([x is None for x in computed_durations]):
            raise RuntimeError(
                "Unable to automatically infer duration of protocol. "
                "Must define stop or duration for at least one procedure"
            )
        return computed_durations[-1]

    def _compile(
        self, dry_run: bool = True, _visualization: bool = False
    ) -> Dict[ActiveComponent, List[Dict[str, Union[float, str, Dict[str, Any]]]]]:
        """
        Compile the protocol into a dict of devices and their procedures.

        Returns:
        - A dict with components as the values and lists of their procedures as the value.
        The elements of the list of procedures are dicts with two keys:
            "time" in seconds
            "params", whose value is a dict of parameters for the procedure.

        Raises:
        - `RuntimeError`: When compilation fails.
        """
        output = {}

        # deal only with compiling active components
        for component in self.apparatus[ActiveComponent]:
            # determine the procedures for each component
            component_procedures: List[MutableMapping] = sorted(
                [x for x in self.procedures if x["component"] == component],
                key=lambda x: x["start"],
            )

            # skip compiling components without procedures
            if not len(component_procedures):
                warn(
                    f"{component} is an active component but was not used in this procedure."
                    " If this is intentional, ignore this warning."
                )
                continue

            # validate each component
            try:
                component._validate(dry_run=dry_run)
            except Exception as e:
                raise RuntimeError(f"{component} isn't valid. Got error: '{str(e)}'.") from e

            # check for conflicting continuous procedures
            if (
                len(
                    [
                        x
                        for x in component_procedures
                        if x["start"] is None and x["stop"] is None
                    ]
                )
                > 1
            ):
                raise RuntimeError(
                    f"{component} cannot have two procedures for the entire duration of the protocol. "
                    "If each procedure defines a different attribute to be set for the entire duration, "
                    "combine them into one call to add(). Otherwise, reduce ambiguity by defining start "
                    "and stop times for each procedure. "
                    ""
                )

            for i, procedure in enumerate(component_procedures):
                # automatically infer start and stop times
                try:
                    # the start time of the next procedure
                    next_start = component_procedures[i + 1]["start"]
                    if next_start == 0:
                        raise RuntimeError(
                            f"Ambiguous start time for {procedure['component']}. "
                        )
                    elif next_start is not None and procedure["stop"] is None:
                        warn(
                            f"Automatically inferring stop time for {procedure['component']} "
                            f"as beginning of {procedure['component']}'s next procedure."
                        )
                        procedure["stop"] = next_start

                    # check for overlapping procedures
                    elif next_start < procedure["stop"] and not isclose(
                        next_start, procedure["stop"]
                    ):
                        msg = "Cannot have two overlapping procedures. "
                        msg += f"{procedure} and {component_procedures[i + 1]} conflict"
                        raise RuntimeError(msg)

                except IndexError:
                    if procedure["stop"] is None:
                        warn(
                            f"Automatically inferring stop for {procedure['component']} as the end of the protocol. "
                            f"To override, provide stop in your call to add()."
                        )
                        procedure["stop"] = self._inferred_duration

            # give the component instructions at all times
            compiled = []
            for i, procedure in enumerate(component_procedures):
                if _visualization:
                    compiled.append(
                        dict(
                            start=procedure["start"],
                            stop=procedure["stop"],
                            params=procedure["params"],
                        )
                    )
                else:
                    compiled.append(
                        dict(time=procedure["start"], params=procedure["params"])
                    )

                    # if the procedure is over at the same time as the next
                    # procedure begins, don't go back to the base state
                    try:
                        if isclose(
                            component_procedures[i + 1]["start"], procedure["stop"]
                        ):
                            continue
                    except IndexError:
                        pass

                    # otherwise, go back to base state
                    new_state = {
                        "time": procedure["stop"],
                        "params": component._base_state,
                    }
                    compiled.append(new_state)

            output[component] = compiled

            # raise warning if duration is explicitly given but not used?
        return output

    def to_dict(self):
        compiled = deepcopy(self._compile(dry_run=True))
        compiled = {k.name: v for (k, v) in compiled.items()}
        return compiled

    def to_list(self):
        output = []
        for procedure in deepcopy(self.procedures):
            procedure["component"] = procedure["component"].name
            output.append(procedure)
        return output

    def yaml(self) -> Union[str, Code]:
        """
        Outputs the uncompiled procedures to YAML.

        Internally, this is a conversion of the output of `Protocol.json` for the purpose of enhanced human readability.

        Returns:
        - YAML of the procedure list.
        When in Jupyter, this string is wrapped in an `IPython.display.Code` object for nice syntax highlighting.

        """
        compiled_yaml = yaml.safe_dump(self.to_list(), default_flow_style=False)

        if get_ipython():
            return Code(compiled_yaml, language="yaml")
        return compiled_yaml

    def json(self) -> Union[str, Code]:
        """
        Outputs the uncompiled procedures to JSON.

        Returns:
        - JSON of the protocol.
          When in Jupyter, this string is wrapped in a `IPython.display.Code` object for nice syntax highlighting.
        """
        compiled_json = json.dumps(self.to_list(), sort_keys=True, indent=4)

        if get_ipython():
            return Code(compiled_json, language="json")
        return compiled_json

    def visualize(self, legend: bool = False, width=500, renderer: str = "notebook"):
        """
        Generates a Gantt plot visualization of the protocol.

        Arguments:
        - `legend`: Whether to show a legend.
        - `renderer`: Which renderer to use. Defaults to "notebook" but can also be "jupyterlab", or "nteract", depending on the development environment. If not in a Jupyter Notebook, this argument is ignored.
        - `width`: The width of the Gantt chart.

        Returns:
        - An interactive visualization of the protocol.
        """

        # don't try to render a visualization to the notebook if we're not in one
        if get_ipython():
            alt.renderers.enable(renderer)

        for component, procedures in self._compile(_visualization=True).items():
            # generate a dict that will be a row in the dataframe
            for procedure in procedures:
                procedure["component"] = str(component)
                procedure["start"] = pd.Timestamp(procedure["start"], unit="s")
                procedure["stop"] = pd.Timestamp(procedure["stop"], unit="s")

                # hoist the params to the main dict
                assert isinstance(procedure["params"], dict)  # needed for typing
                for k, v in procedure["params"].items():
                    procedure[k] = v

                # show what the valve is actually connecting to
                if isinstance(component, Valve) and type(procedure["setting"]) == int:
                    assert isinstance(component.mapping, Mapping)
                    # guess the component, c, which the valve is set to
                    mapped_component = [
                        repr(k)
                        for k, v in component.mapping.items()
                        if v == procedure["setting"]
                    ][0]
                    procedure["mapped component"] = mapped_component
                # TODO: make this deterministic for color coordination
                procedure["params"] = json.dumps(procedure["params"])

            # prettyify the tooltips
            tooltips = [
                alt.Tooltip("utchoursminutesseconds(start):T", title="start (h:m:s)"),
                alt.Tooltip("utchoursminutesseconds(stop):T", title="stop (h:m:s)"),
                "component",
            ]

            # just add the params to the tooltip
            tooltips.extend(
                [
                    x
                    for x in procedures[0].keys()
                    if x not in ["component", "start", "stop", "params"]
                ]
            )

            # generate the component's graph
            source = pd.DataFrame(procedures)
            component_chart = (
                alt.Chart(source, width=width)
                .mark_bar()
                .encode(
                    x="utchoursminutesseconds(start):T",
                    x2="utchoursminutesseconds(stop):T",
                    y="component",
                    color=alt.Color("params:N", legend=None)
                    if not legend
                    else "params",
                    tooltip=tooltips,
                )
            )

            # label the axes
            component_chart.encoding.x.title = "Experiment Elapsed Time (h:m:s)"
            component_chart.encoding.y.title = "Component"

            # combine with the other charts
            try:
                chart += component_chart  # type: ignore
            except NameError:
                chart = component_chart

        return chart.interactive()

    def execute(
        self,
        dry_run: Union[bool, int] = False,
        verbosity: str = "info",
        confirm: bool = False,
        strict: bool = True,
        log_file: Union[str, bool, os.PathLike, None] = True,
        log_file_verbosity: Optional[str] = "trace",
        log_file_compression: Optional[str] = None,
        data_file: Union[str, bool, os.PathLike, None] = True,
    ) -> Experiment:
        """
        Executes the procedure.

        Arguments:
        - `confirm`: Whether to bypass the manual confirmation message before execution.
        - `dry_run`: Whether to simulate the experiment or actually perform it. Defaults to `False`,
        which means executing the protocol on real hardware. If an integer greater than zero,
        the dry run will execute at that many times speed.
        - `strict`: Whether to stop execution upon encountering any errors.
        If False, errors will be noted but ignored.
        - `verbosity`: The level of logging verbosity. One of "critical", "error", "warning", "success", "info", "debug", or "trace" in descending order of severity. "debug" and (especially) "trace" are not meant to be used regularly, as they generate significant amounts of usually useless information. However, these verbosity levels are useful for tracing where exactly a bug was generated, especially if no error message was thrown.
        - `log_file`: The file to write the logs to during execution. If `True`, the data will be written to a file in `~/.mechwolf` with the filename `{experiment_id}.log.jsonl`. If falsey, no logs will be written to the file.
        - `log_file_verbosity`: How verbose the logs in file should be. By default, it is "trace", which is the most verbose logging available. If `None`, it will use the same level as `verbosity`.
        - `log_file_compression`: Whether to compress the log file after the experiment.
        - `data_file`: The file to write the experimental data to during execution. If `True`, the data will be written to a file in `~/.mechwolf` with the filename `{experiment_id}.data.jsonl`. If falsey, no data will be written to the file.

        Returns:
        - An `Experiment` object. In a Jupyter notebook, the object yields an interactive visualization. If protocol execution fails for any reason that does not raise an error, the return type is None.

        Raises:
        - `RuntimeError`: When attempting to execute a protocol on invalid components.
        """

        # the Experiment object is going to hold all the info
        E = Experiment(self)
        E._execute(
            dry_run=dry_run,
            verbosity=verbosity,
            confirm=confirm,
            strict=strict,
            log_file=log_file,
            log_file_verbosity=log_file_verbosity,
            log_file_compression=log_file_compression,
            data_file=data_file,
        )

        return E



"""
Knauer pump control.
"""
import asyncio
import warnings
from loguru import logger
from enum import Enum

from flowchem.components.devices.Knauer.Knauer_common import KnauerEthernetDevice
from flowchem.components.stdlib import Pump
from flowchem.exceptions import DeviceError
from flowchem.units import flowchem_ureg

FLOW = "FLOW"  # 0-50000 µL/min, int only!
HEADTYPE = "HEADTYPE"  # 10, 50 ml. Value refers to highest flowrate in ml/min
PMIN10 = "PMIN10"  # 0-400 in 0.1 MPa, use to avoid dryrunning
PMIN50 = "PMIN50"  # 0-150 in 0.1 MPa, use to avoid dryrunning
PMAX10 = "PMAX10"  # 0-400 in 0.1 MPa, chosen automatically by selecting pump head
PMAX50 = "PMAX50"  # 0-150 in 0.1 MPa, chosen automatically by selecting pumphead
IMIN10 = "IMIN10"  # 0-100 minimum motor current
IMIN50 = "IMIN50"  # 0-100 minimum motor current
STARTLEVEL = "STARTLEVEL"  # 0, 1 configures start. 0 -> only start pump when shorted to GND, 1 -> always allow start
ERRIO = "ERRIO"  # 0, 1 write/read error in/output ??? sets errio either 1 or 0, reports errio:ok
STARTMODE = "STARTMODE"  # 0, 1; 0=pause pump after switchon, 1=start immediatley with previous set flow rate
ADJ10 = "ADJ10"  # 100-2000
ADJ50 = "ADJ50"  # 100-2000
CORR10 = "CORR10"  # 0-300
CORR50 = "CORR50"  # 0-300
EXTFLOW = "EXTFLOW?"
IMOTOR = "IMOTOR"  # motor current in relative units 0-100
PRESSURE = "PRESSURE?"  # reads the pressure in 0.1 MPa
ERRORS = "ERRORS"  # displays last 5 error codes
EXTCONTR = "EXTCONTR"  # allows flow control via analog input
LOCAL = "LOCAL"  # no parameter, releases pump to manual control
REMOTE = "REMOTE"  # manual param input prevented
PUMP_ON = "ON"  # starts flow
PUMP_OFF = "OFF"  # stops flow


class AzuraPumpHeads(Enum):
    """Two Pump heads are available for the Azura: 50 mL/min and 10 mL/min."""

    FLOWRATE_FIFTY_ML = 50
    FLOWRATE_TEN_ML = 10


class AzuraCompactPump(KnauerEthernetDevice, Pump):
    """Control module for Knauer Azura Compact pumps."""

    metadata = {
        "author": [
            {
                "first_name": "Jakob",
                "last_name": "Wolf",
                "email": "jakob.wolf@mpikg.mpg.de",
                "institution": "Max Planck Institute of Colloids and Interfaces",
                "github_username": "JB-Wolf",
            },
            {
                "first_name": "Dario",
                "last_name": "Cambie",
                "email": "dario.cambie@mpikg.mpg.de",
                "institution": "Max Planck Institute of Colloids and Interfaces",
                "github_username": "dcambie",
            },
        ],
        "stability": "beta",
        "supported": True,
    }

    def __init__(self, ip_address=None, mac_address=None, name=None, max_pressure: str = None):
        super().__init__(ip_address, mac_address, name)
        self.eol = b"\n\r"

        # All of the following are set upon initialize()
        self.max_allowed_pressure, self.max_allowed_flow = 0, 0
        self._headtype = None
        self._running = None
        self._pressure_limit = max_pressure

        self.rate = flowchem_ureg.parse_expression("0 ml/min")
        self._visualization_shape = "box3d"
        self._base_state = dict(rate="0 mL/min")

    async def initialize(self):
        """Initialize connection"""
        # Here the magic happens...
        await super().initialize()

        # Here it is checked that the device is a pump and not a valve
        await self.get_headtype()
        # Place pump in remote control
        await self.set_remote()
        # Also ensure rest state is not pumping.
        await self.stop_flow()

        if self._pressure_limit is not None:
            await self.set_maximum_pressure(self._pressure_limit)

    @staticmethod
    def error_present(reply: str) -> bool:
        """True if there are errors, False otherwise. Warns for errors."""

        if not reply.startswith("ERROR"):
            return False

        if "ERROR:1" in reply:
            warnings.warn("Invalid message sent to device.\n")

        elif "ERROR:2" in reply:
            warnings.warn(
                "Setpoint refused by device.\n"
                "Refer to manual for allowed values.\n"
            )
        else:
            warnings.warn("Unspecified error detected!")
        return True

    async def _transmit_and_parse_reply(self, message: str) -> str:
        """
        sends command and receives reply, deals with all communication based stuff and checks
        that the valve is of expected type
        :param message:
        :return: reply: str
        """
        reply = await self._send_and_receive(message)
        if self.error_present(reply):
            return ""

        # Setpoint ok
        elif ":OK" in reply:
            logger.debug("setpoint successfully set!")
            return "OK"

        # Replies to 'VALUE?' are in the format 'VALUE:'
        elif message[:-1] in reply:
            logger.debug(f"setpoint successfully acquired, value={reply}")
            # Last value after colon
            return reply.split(":")[-1]

        # No reply
        elif not reply:
            warnings.warn("No reply received")
            return ""

        warnings.warn(f"Unrecognized reply: {reply}")
        return reply

    async def create_and_send_command(
        self, message, setpoint: int = None, setpoint_range: tuple = None
    ):
        """
        Create and sends a message from the command.

        If setpoint is given, then the command is appended with :value
        If not setpoint is given, a "?" is added for getter syntax

        e.g. message = "HEADTYPE"
        no setpoint -> sends "HEADTYPE?"
        w/ setpoint -> sends "HEADTYPE:<setpoint_value>"
        """
        # GETTER
        if setpoint is None:
            return await self._transmit_and_parse_reply(message + "?")

        # SETTER with range
        if setpoint_range:
            if setpoint in range(*setpoint_range):
                return await self._transmit_and_parse_reply(
                    message + ":" + str(setpoint)
                )

            warnings.warn(
                f"The setpoint provided {setpoint} is not valid for the command "
                f"{message}!\n Accepted range is: {setpoint_range}.\n"
                f"Command ignored"
            )
            return ""

        # SETTER w/o range
        else:
            return await self._transmit_and_parse_reply(message + ":" + str(setpoint))

    @property
    def _headtype(self):
        """Internal state reflecting pump one, use set_headtype() to change in pump!"""
        return self.__headtype

    @_headtype.setter
    def _headtype(self, htype):
        self.__headtype = htype

        if htype == AzuraPumpHeads.FLOWRATE_TEN_ML:
            self.max_allowed_pressure, self.max_allowed_flow = 400, 10000
        elif htype == AzuraPumpHeads.FLOWRATE_FIFTY_ML:
            self.max_allowed_pressure, self.max_allowed_flow = 150, 50000

    async def get_headtype(self):
        """Returns pump's head type."""
        head_type_id = await self.create_and_send_command(HEADTYPE)
        try:
            headtype = AzuraPumpHeads(int(head_type_id))
            # Sets internal property (changes max flowrate etc)
            self._headtype = headtype
        except ValueError as e:
            raise DeviceError(
                "It seems you're trying instantiate an unknown device/unknown pump type as Knauer Pump.\n"
                "Only Knauer Azura Compact is supported"
            ) from e
        logger.debug(f"Head type of pump {self.ip_address} is {headtype}")

        return headtype

    async def set_headtype(self, head_type: AzuraPumpHeads):
        """Sets pump's head type."""
        await self.create_and_send_command(HEADTYPE, setpoint=head_type.value)
        # Update internal property (changes max flowrate etc)
        self._headtype = head_type
        logger.debug(f"Head type set to {head_type}")

    async def get_flow(self) -> str:
        """ Gets flow rate. """
        flow_value = await self.create_and_send_command(FLOW)
        flowrate = flowchem_ureg(f"{flow_value} ul/min")
        logger.debug(f"Current flow rate is {flowrate}")
        return str(flowrate.to("ml/min"))

    async def set_flow(self, flowrate: str = None):
        """ Sets flow rate.

        :param flowrate: string with units
        """
        parsed_flowrate = flowchem_ureg(flowrate)
        await self.create_and_send_command(
            FLOW,
            setpoint=round(parsed_flowrate.m_as("ul/min")),
            setpoint_range=(0, self.max_allowed_flow + 1),
        )
        logger.info(f"Flow set to {flowrate}")

    async def get_minimum_pressure(self):
        """Gets minimum pressure. The pumps stops if the measured P is lower than this."""

        command = PMIN10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else PMIN50
        p_min = await self.create_and_send_command(command) * flowchem_ureg.bar
        return str(p_min)

    async def set_minimum_pressure(self, value: str = "0 bar"):
        """ Sets minimum pressure. The pumps stops if the measured P is lower than this. """

        pressure = flowchem_ureg(value)
        command = PMIN10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else PMIN50
        await self.create_and_send_command(
            command,
            setpoint=round(pressure.m_as("bar")),
            setpoint_range=(0, self.max_allowed_pressure + 1),
        )
        logger.info(f"Minimum pressure set to {pressure}")

    async def get_maximum_pressure(self) -> str:
        """Gets maximum pressure. The pumps stops if the measured P is higher than this."""

        command = PMAX10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else PMAX50
        p_max = await self.create_and_send_command(command) * flowchem_ureg.bar
        return str(p_max)

    async def set_maximum_pressure(self, value: str):
        """ Sets maximum pressure. The pumps stops if the measured P is higher than this. """

        pressure = flowchem_ureg(value)
        command = PMAX10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else PMAX50
        await self.create_and_send_command(
            command,
            setpoint=round(pressure.m_as("bar")),
            setpoint_range=(0, self.max_allowed_pressure + 1),
        )
        logger.info(f"Maximum pressure set to {pressure}")

    async def set_minimum_motor_current(self, setpoint=None):
        """Sets minimum motor current."""
        command = IMIN10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else IMIN50

        reply = await self.create_and_send_command(
            command, setpoint=setpoint, setpoint_range=(0, 101)
        )
        logger.debug(f"Minimum motor current set to {setpoint}, returns {reply}")

    async def is_start_in_required(self):
        """
        Check state of START IN. See require_start_in() for details.
        """
        runlevel = await self.create_and_send_command(STARTLEVEL)
        return not bool(int(runlevel))

    async def require_start_in(self, value: bool = True):
        """
        Configures START IN. If required, the pump starts only if the STARTIN pin is shortened to GND.

        True = Pump starts the flow at short circuit contact only. (Start In <> Ground). [0]
        False = Pump starts the flow without a short circuit contact. (Start In <> Ground). [1]
        """
        setpoint = int(not value)
        await self.create_and_send_command(STARTLEVEL, setpoint=setpoint)
        logger.debug(f"Start in required set to {value}")

    async def is_autostart_enabled(self):
        """Returns the default behaviour of the pump upon power on."""
        reply = await self.create_and_send_command(STARTMODE)
        return bool(int(reply))

    async def enable_autostart(self, value: bool = True):
        """
        Sets the default behaviour of the pump upon power on.

        :param value: False: pause pump after switch on. True: start pumping with previous flow rate at startup
        :return: device message
        """
        await self.create_and_send_command(STARTMODE, setpoint=int(value))
        logger.debug(f"Autostart set to {value}")

    async def get_adjusting_factor(self):
        """Gets the adjust parameter. Not clear what it is."""
        command = ADJ10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else ADJ50
        reply = await self.create_and_send_command(command)
        return int(reply)

    async def set_adjusting_factor(self, setpoint: int = None):
        """Sets the adjust parameter. Not clear what it is."""
        command = ADJ10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else ADJ50
        reply = await self.create_and_send_command(
            command, setpoint=setpoint, setpoint_range=(0, 2001)
        )
        logger.debug(f"Adjusting factor of set to {setpoint}, returns {reply}")

    async def get_correction_factor(self):
        """Gets the correction factor. Not clear what it is."""
        command = CORR10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else CORR50
        return int(await self.create_and_send_command(command))

    async def set_correction_factor(self, setpoint=None):
        """Sets the correction factor. Not clear what it is."""
        command = CORR10 if self._headtype == AzuraPumpHeads.FLOWRATE_TEN_ML else CORR50
        reply = await self.create_and_send_command(
            command, setpoint=setpoint, setpoint_range=(0, 301)
        )
        logger.debug(f"Correction factor set to {setpoint}, returns {reply}")

    async def read_pressure(self) -> str:
        """ If the pump has a pressure sensor, returns pressure. Read-only property of course. """
        pressure = await self._transmit_and_parse_reply(PRESSURE) * flowchem_ureg.bar
        logger.debug(f"Pressure measured = {pressure}")
        return str(pressure)

    async def read_extflow(self) -> float:
        """Read the set flowrate from analog in."""
        ext_flow = await self._transmit_and_parse_reply(EXTFLOW)
        logger.debug(f"Extflow reading returns {ext_flow}")
        return float(ext_flow)

    async def read_errors(self):
        """Returns the last 5 errors."""
        last_5_errors = await self.create_and_send_command(ERRORS)
        logger.debug(f"Error reading returns {last_5_errors}")
        return last_5_errors

    async def read_motor_current(self):
        """Returns motor current, relative in percent 0-100."""
        current_percent = int(await self.create_and_send_command(IMOTOR))
        logger.debug(f"Motor current reading returns {current_percent} %")
        return current_percent

    async def start_flow(self):
        """Starts flow"""
        await self._transmit_and_parse_reply(PUMP_ON)
        self._running = True
        logger.info("Pump switched on")

    async def stop_flow(self):
        """Stops flow"""
        await self._transmit_and_parse_reply(PUMP_OFF)
        self._running = False
        logger.info("Pump not pumping")

    def is_running(self):
        """Get pump state."""
        return self._running

    async def set_local(self, state: bool = True):
        """Relinquish remote control"""
        await self.create_and_send_command(LOCAL, setpoint=int(state))
        logger.debug(f"Local control set to {state}")

    async def set_remote(self, state: bool = True):
        """Set remote control on or off."""
        await self.create_and_send_command(REMOTE, setpoint=int(state))
        logger.debug(f"Remote control set to {state}")

    async def set_errio(self, param: bool):
        """no idea what this exactly does..."""
        await self.create_and_send_command(ERRIO, setpoint=int(param))
        logger.debug(f"Set errio {param}")

    async def is_analog_control_enabled(self):
        """Returns the status of the external flow control via analog input."""
        reply = await self.create_and_send_command(EXTCONTR)
        return bool(int(reply))

    async def enable_analog_control(self, value: bool):
        """External flow control via analog input.

        False = prevents external flow control. [0]
        True = allows the flow rate control via analog input 0 - 10V (10ml: 1 V = 1 ml/min, 50ml: 1 V = 5 ml/min). [1]
        """
        await self.create_and_send_command(EXTCONTR, setpoint=int(value))
        logger.debug(f"External control set to {value}")

    async def __aenter__(self):
        await self.initialize()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.stop_flow()

    async def _update(self):
        """Called automatically to change flow rate."""

        if self.rate == 0:
            await self.stop_flow()
        else:
            await self.set_flow(self.rate)
            await self.start_flow()


if __name__ == "__main__":
    # This is a bug of asyncio on Windows :|
    import sys

    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    p = AzuraCompactPump(ip_address="192.168.10.113")

    async def main(pump: AzuraCompactPump):
        """Test function"""
        await pump.initialize()
        await pump.set_flow("0.1 ml/min")
        await pump.start_flow()
        await asyncio.sleep(5)
        await pump.stop_flow()

    asyncio.run(main(p))
